import dataclasses
import heapq
from typing import Generator, List, Tuple

try:
    import pynamodb  # type: ignore
except ImportError:
    raise RuntimeError(
        "pynamodb is not detected in current environment, install Wicker with extra arguments:"
        " `pip install wicker[dynamodb]`"
    )
from pynamodb.attributes import NumberAttribute, UnicodeAttribute
from pynamodb.models import Model
from retry import retry

from wicker.core.config import get_config
from wicker.core.definitions import DatasetID
from wicker.core.writer import (
    AbstractDatasetWriterMetadataDatabase,
    ExampleKey,
    MetadataDatabaseScanRow,
)

# DANGER: If these constants are ever changed, this is a backward-incompatible change.
# Make sure that all the writers and readers of dynamodb are in sync when changing.
NUM_DYNAMODB_SHARDS = 32
HASH_PREFIX_LENGTH = 4
DYNAMODB_QUERY_PAGINATION_LIMIT = 1000


@dataclasses.dataclass(frozen=True)
class DynamoDBConfig:
    table_name: str
    region: str


def get_dynamodb_config() -> DynamoDBConfig:
    raw_config = get_config().raw
    if "dynamodb_config" not in raw_config:
        raise RuntimeError("Could not find 'dynamodb' parameters in config")
    if "table_name" not in raw_config["dynamodb_config"]:
        raise RuntimeError("Could not find 'table_name' parameter in config.dynamodb_config")
    if "region" not in raw_config["dynamodb_config"]:
        raise RuntimeError("Could not find 'region' parameter in config.dynamodb_config")
    return DynamoDBConfig(
        table_name=raw_config["dynamodb_config"]["table_name"],
        region=raw_config["dynamodb_config"]["region"],
    )


class DynamoDBExampleDBRow(Model):
    class Meta:
        table_name = get_dynamodb_config().table_name
        region = get_dynamodb_config().region

    dataset_id = UnicodeAttribute(hash_key=True)
    example_id = UnicodeAttribute(range_key=True)
    partition = UnicodeAttribute()
    row_data_path = UnicodeAttribute()
    row_size = NumberAttribute()


def _key_to_row_id_and_shard_id(example_key: ExampleKey) -> Tuple[str, int]:
    """Deterministically convert an ExampleKey into a row_id and a shard_id which are used as
    DynamoDB RangeKeys and HashKeys respectively.

    HashKeys help to increase read/write throughput by allowing us to use different partitions.
    RangeKeys are how the rows are sorted within partitions by DynamoDB.

    We completely randomize the hash and range key to optimize for write throughput, but this means
    that sorting needs to be done entirely client-side in our application.
    """
    partition_example_id = f"{example_key.partition}//{'/'.join([str(key) for key in example_key.primary_key_values])}"
    hash = example_key.hash()
    shard = int(hash, 16) % NUM_DYNAMODB_SHARDS
    return partition_example_id, shard


def _dataset_shard_name(dataset_id: DatasetID, shard_id: int) -> str:
    """Get the name of the DynamoDB partition for a given dataset_definition and shard number"""
    return f"{dataset_id}_shard{shard_id:02d}"


class DynamodbMetadataDatabase(AbstractDatasetWriterMetadataDatabase):
    def save_row_metadata(self, dataset_id: DatasetID, key: ExampleKey, location: str, row_size: int) -> None:
        """Saves a row in the metadata database, marking it as having been uploaded to S3 and
        ready for shuffling.

        :param dataset_id: The ID of the dataset to save to
        :param key: The key of the example
        :param location: The location of the example in S3
        :param row_size: The size of the file in S3
        """
        partition_example_id, shard_id = _key_to_row_id_and_shard_id(key)
        entry = DynamoDBExampleDBRow(
            dataset_id=_dataset_shard_name(dataset_id, shard_id),
            example_id=partition_example_id,
            partition=key.partition,
            row_data_path=location,
            row_size=row_size,
        )
        entry.save()

    def scan_sorted(self, dataset_id: DatasetID) -> Generator[MetadataDatabaseScanRow, None, None]:
        """Scans the MetadataDatabase for a **SORTED** list of ExampleKeys for a given dataset. Should be fast O(minutes)
        to perform as this will be called from a single machine to assign chunks to jobs to run.

        :param dataset: The dataset to scan the metadata database for
        :return: a Generator of MetadataDatabaseScanRow in **SORTED** primary_key order
        """

        @retry(pynamodb.exceptions.QueryError, tries=10, backoff=2, delay=4, jitter=(0, 2))
        def shard_iterator(shard_id: int) -> Generator[DynamoDBExampleDBRow, None, None]:
            """Yields DynamoDBExampleDBRows from a given shard to exhaustion, sorted by example_id in ascending order
            DynamoDBExampleDBRows are yielded in sorted order of the Dynamodb RangeKey, which is the example_id
            """
            hash_key = _dataset_shard_name(dataset_id, shard_id)
            last_evaluated_key = None
            while True:
                query_results = DynamoDBExampleDBRow.query(
                    hash_key,
                    consistent_read=True,
                    last_evaluated_key=last_evaluated_key,
                    limit=DYNAMODB_QUERY_PAGINATION_LIMIT,
                )
                yield from query_results
                if query_results.last_evaluated_key is None:
                    break
                last_evaluated_key = query_results.last_evaluated_key

        # Individual shards have their rows already in sorted order
        # Elements are popped off each shard to exhaustion and put into a minheap
        # We yield from the heap to exhaustion to provide a stream of globally sorted example_ids
        heap: List[Tuple[str, int, DynamoDBExampleDBRow]] = []
        shard_iterators = {shard_id: shard_iterator(shard_id) for shard_id in range(NUM_DYNAMODB_SHARDS)}
        for shard_id, iterator in shard_iterators.items():
            try:
                row = next(iterator)
                heapq.heappush(heap, (row.example_id, shard_id, row))
            except StopIteration:
                pass
        while heap:
            _, shard_id, row = heapq.heappop(heap)
            try:
                nextrow = next(shard_iterators[shard_id])
                heapq.heappush(heap, (nextrow.example_id, shard_id, nextrow))
            except StopIteration:
                pass
            yield MetadataDatabaseScanRow(
                partition=row.partition,
                row_data_path=row.row_data_path,
                row_size=row.row_size,
            )



import os
import tempfile
from unittest import TestCase, mock

from botocore.exceptions import ClientError  # type: ignore
from botocore.stub import Stubber  # type: ignore

from wicker.core.storage import S3DataStorage


class TestS3DataStorage(TestCase):
    def test_bucket_key_from_s3_path(self) -> None:
        """Unit test for the S3DataStorage bucket_key_from_s3_path function"""
        data_storage = S3DataStorage()

        s3_url = "s3://hello/world"
        bucket, key = data_storage.bucket_key_from_s3_path(s3_url)
        self.assertEqual(bucket, "hello")
        self.assertEqual(key, "world")

        s3_url = "s3://hello/"
        bucket, key = data_storage.bucket_key_from_s3_path(s3_url)
        self.assertEqual(bucket, "hello")
        self.assertEqual(key, "")

        s3_url = "s3://"
        bucket, key = data_storage.bucket_key_from_s3_path(s3_url)
        self.assertEqual(bucket, "")
        self.assertEqual(key, "")

        s3_url = "s3://hello/world/"
        bucket, key = data_storage.bucket_key_from_s3_path(s3_url)
        self.assertEqual(bucket, "hello")
        self.assertEqual(key, "world/")

    def test_check_exists_s3(self) -> None:
        """Unit test for the check_exists_s3 function."""
        data_storage = S3DataStorage()
        input_path = "s3://foo/bar/baz/dummy"

        with Stubber(data_storage.client) as stubber:
            response = {}  # type: ignore
            expected_params = {"Bucket": "foo", "Key": "bar/baz/dummy"}
            stubber.add_response("head_object", response, expected_params)
            self.assertTrue(data_storage.check_exists_s3(input_path))

    def test_check_exists_s3_nonexisting(self) -> None:
        """Unit test for the check_exists_s3 function."""
        data_storage = S3DataStorage()
        input_path = "s3://foo/bar/baz/dummy"

        with Stubber(data_storage.client) as stubber:
            stubber.add_client_error(
                expected_params={"Bucket": "foo", "Key": "bar/baz/dummy"},
                method="head_object",
                service_error_code="404",
                service_message="The specified key does not exist.",
            )

            # The check_exists_s3 function catches the exception when the key does not exist
            self.assertFalse(data_storage.check_exists_s3(input_path))

    def test_put_object_s3(self) -> None:
        """Unit test for the put_object_s3 function."""
        data_storage = S3DataStorage()
        object_bytes = b"this is my object"
        input_path = "s3://foo/bar/baz/dummy"

        with Stubber(data_storage.client) as stubber:
            response = {}  # type: ignore
            expected_params = {
                "Body": object_bytes,
                "Bucket": "foo",
                "Key": "bar/baz/dummy",
            }
            stubber.add_response("put_object", response, expected_params)
            data_storage.put_object_s3(object_bytes, input_path)

    def test_put_file_s3(self) -> None:
        """Unit test for the put_file_s3 function"""
        data_storage = S3DataStorage()
        object_bytes = b"this is my object"
        input_path = "s3://foo/bar/baz/dummy"

        with tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.write(object_bytes)
            tmpfile.flush()

            with Stubber(data_storage.client) as stubber:
                response = {}  # type: ignore
                stubber.add_response("put_object", response, None)
                data_storage.put_file_s3(tmpfile.name, input_path)

    @staticmethod
    def download_file_side_effect(*args, **kwargs) -> None:  # type: ignore
        """Helper function to patch the S3 download_file function with a side-effect that creates an
        empty file at the correct path in order to mock the download"""
        input_path = str(kwargs["filename"])
        with open(input_path, "w"):
            pass

    # Stubber does not have a stub function for S3 client download_file function, so patch it
    @mock.patch("boto3.s3.transfer.S3Transfer.download_file")
    def test_fetch_file_s3(self, download_file: mock.Mock) -> None:
        """Unit test for the fetch_file_s3 function."""
        data_storage = S3DataStorage()
        input_path = "s3://foo/bar/baz/dummy"
        with tempfile.TemporaryDirectory() as local_prefix:
            # Add a side-effect to create the file to download at the correct local path
            download_file.side_effect = self.download_file_side_effect

            local_path = data_storage.fetch_file_s3(input_path, local_prefix)
            download_file.assert_called_once_with(
                bucket="foo",
                key="bar/baz/dummy",
                filename=f"{local_prefix}/bar/baz/dummy",
                extra_args=None,
                callback=None,
            )
            self.assertTrue(os.path.isfile(local_path))

    # Stubber does not have a stub function for S3 client download_file function, so patch it
    @mock.patch("boto3.s3.transfer.S3Transfer.download_file")
    def test_fetch_file_s3_on_nonexistent_file(self, download_file: mock.Mock) -> None:
        """Unit test for the fetch_file_s3 function for a non-existent file in S3."""
        data_storage = S3DataStorage()
        input_path = "s3://foo/bar/barbazz/dummy"
        local_prefix = "/tmp"

        response = {"Error": {"Code": "404"}}
        side_effect = ClientError(response, "unexpected")
        download_file.side_effect = side_effect

        with self.assertRaises(ClientError):
            data_storage.fetch_file_s3(input_path, local_prefix)



import json
import os
from unittest.mock import call, patch

import pytest

from wicker.plugins.wandb import (
    _identify_s3_url_for_dataset_version,
    _set_wandb_credentials,
    version_dataset,
)


@pytest.fixture
def temp_config(request, tmpdir):
    # parse the config json into a proper config
    config_json = request.param.get("config_json")
    parsable_config = {
        "wandb_config": config_json.get("wandb_config", {}),
        "aws_s3_config": {
            "s3_datasets_path": config_json.get("aws_s3_config", {}).get("s3_datasets_path", None),
            "region": config_json.get("aws_s3_config", {}).get("region", None),
        },
    }

    # set the config path env and then get the creds from the function
    temp_config = tmpdir / "temp_config.json"
    with open(temp_config, "w") as file_stream:
        json.dump(parsable_config, file_stream)
    os.environ["WICKER_CONFIG_PATH"] = str(temp_config)
    return temp_config, parsable_config


@pytest.mark.parametrize(
    "temp_config, dataset_name, dataset_version, dataset_metadata, entity",
    [
        (
            {
                "config_json": {
                    "aws_s3_config": {
                        "s3_datasets_path": "s3://test_path/to_nowhere/",
                        "region": "us-west-2",
                    },
                    "wandb_config": {
                        "wandb_api_key": "test_key",
                        "wandb_base_url": "test_url",
                    },
                }
            },
            "test_data",
            "0.0.1",
            {"test_key": "test_value"},
            "test_entity",
        )
    ],
    indirect=["temp_config"],
)
def test_version_dataset(temp_config, dataset_name, dataset_version, dataset_metadata, entity):
    """
    GIVEN: A mocked dataset folder to track, a dataset version to track, metadata to track, and a backend to use
    WHEN: This dataset is registered or versioned on W&B
    THEN: The dataset shows up as a run on the portal
    """
    # need to mock out all the wandb calls and test just the inputs to them
    _, config = temp_config
    with patch("wicker.plugins.wandb.wandb.init") as patched_wandb_init:
        with patch("wicker.plugins.wandb.wandb.Artifact") as patched_artifact:
            # version the dataset with the patched functions/classes
            version_dataset(dataset_name, dataset_version, entity, dataset_metadata)

            # establish the expected calls
            expected_artifact_calls = [
                call(f"{dataset_name}_{dataset_version}", type="dataset"),
                call().add_reference(
                    f"{config['aws_s3_config']['s3_datasets_path']}{dataset_name}/{dataset_version}/assets",
                    name="dataset",
                ),
                call().metadata.__setitem__("version", dataset_version),
                call().metadata.__setitem__(
                    "s3_uri", f"{config['aws_s3_config']['s3_datasets_path']}{dataset_name}/{dataset_version}/assets"
                ),
            ]
            for key, value in dataset_metadata.items():
                expected_artifact_calls.append(call().metadata.__setitem__(key, value))

            expected_run_calls = [
                call(project="dataset_curation", name=f"{dataset_name}_{dataset_version}", entity=entity),
                call().log_artifact(patched_artifact()),
            ]

            # assert that these are properly called
            patched_artifact.assert_has_calls(expected_artifact_calls, any_order=True)

            patched_wandb_init.assert_has_calls(expected_run_calls, any_order=True)


@pytest.mark.parametrize(
    "credentials_to_load, temp_config",
    [
        (
            {"WANDB_BASE_URL": "env_base", "WANDB_API_KEY": "env_key", "WANDB_USER_EMAIL": "env_email"},
            {
                "config_json": {
                    "wandb_config": {
                        "wandb_base_url": "config_base",
                        "wandb_api_key": "config_key",
                    }
                }
            },
        )
    ],
    ids=["basic test to override all params"],
    indirect=["temp_config"],
)
def test_set_wandb_credentials(credentials_to_load, temp_config, tmpdir):
    """
    GIVEN: A set of credentials as existing envs and a config json specifying creds
    WHEN: The configs are requested for wandb
    THEN: The proper env variables should be set in the environment based on rules, default to
          wicker config and reject preset env variables
    """
    temp_config_pth, config_json = temp_config

    # load the creds into the env as the base comparison
    for key, value in credentials_to_load.items():
        os.environ[key] = value

    # compare the creds for expected results
    _set_wandb_credentials()
    assert os.environ["WANDB_BASE_URL"] == config_json["wandb_config"]["wandb_base_url"]
    assert os.environ["WANDB_API_KEY"] == config_json["wandb_config"]["wandb_api_key"]


@pytest.mark.parametrize(
    "temp_config, dataset_name, dataset_version, dataset_backend, expected_url",
    [
        (
            {"config_json": {"aws_s3_config": {"s3_datasets_path": "s3://test_path_to_nowhere", "region": "test"}}},
            "test_dataset",
            "0.0.0",
            "s3",
            "s3://test_path_to_nowhere/test_dataset/0.0.0/assets",
        )
    ],
    ids=["Basic test with s3"],
    indirect=["temp_config"],
)
def test_identify_s3_url_for_dataset_version(temp_config, dataset_name, dataset_version, dataset_backend, expected_url):
    """
    GIVEN: A temporary config, a dataset name, version, backend, and expected url
    WHEN: The assets url is pulled from the path factory with these parameters
    THEN: The expected url should match what is returned from the function
    """
    parsed_dataset_url = _identify_s3_url_for_dataset_version(dataset_name, dataset_version, dataset_backend)
    assert parsed_dataset_url == expected_url



import pickle
import tempfile
import unittest
import unittest.mock
from typing import IO, Dict, List, Tuple

from wicker.core.column_files import ColumnBytesFileLocationV1
from wicker.core.config import get_config
from wicker.core.definitions import DatasetDefinition, DatasetID, DatasetPartition
from wicker.core.shuffle import ShuffleJob, ShuffleJobFactory, ShuffleWorker
from wicker.core.writer import MetadataDatabaseScanRow
from wicker.schema import dataparsing, schema, serialization
from wicker.testing.codecs import Vector, VectorCodec
from wicker.testing.storage import FakeS3DataStorage

DATASET_NAME = "dataset1"
DATASET_VERSION = "0.0.1"
FAKE_DATASET_SCHEMA = schema.DatasetSchema(
    fields=[
        schema.IntField("timestamp"),
        schema.StringField("car_id"),
        schema.ObjectField("vector", VectorCodec(0)),
    ],
    primary_keys=["car_id", "timestamp"],
)
FAKE_EXAMPLE = {
    "timestamp": 1,
    "car_id": "somecar",
    "vector": Vector([0, 0, 0]),
}
FAKE_DATASET_ID = DatasetID(name=DATASET_NAME, version=DATASET_VERSION)
FAKE_DATASET_DEFINITION = DatasetDefinition(
    dataset_id=FAKE_DATASET_ID,
    schema=FAKE_DATASET_SCHEMA,
)


class TestShuffleJobFactory(unittest.TestCase):
    def test_shuffle_job_factory_no_entries(self) -> None:
        """Tests the functionality of a shuffle_job_factory when provided with a mock backend"""
        mock_rows: List[MetadataDatabaseScanRow] = []
        mock_dataset_writer_backend = unittest.mock.MagicMock()
        mock_dataset_writer_backend._metadata_db.scan_sorted.return_value = (row for row in mock_rows)
        job_factory = ShuffleJobFactory(mock_dataset_writer_backend)
        self.assertEqual(list(job_factory.build_shuffle_jobs(FAKE_DATASET_ID)), [])

    def test_shuffle_job_factory_one_entry(self) -> None:
        """Tests the functionality of a shuffle_job_factory when provided with a mock backend"""
        mock_rows: List[MetadataDatabaseScanRow] = [
            MetadataDatabaseScanRow(partition="train", row_data_path="somepath", row_size=1337)
        ]
        mock_dataset_writer_backend = unittest.mock.MagicMock()
        mock_dataset_writer_backend._metadata_db.scan_sorted.return_value = (row for row in mock_rows)
        job_factory = ShuffleJobFactory(mock_dataset_writer_backend)
        self.assertEqual(
            list(job_factory.build_shuffle_jobs(FAKE_DATASET_ID)),
            [
                ShuffleJob(
                    dataset_partition=DatasetPartition(
                        dataset_id=FAKE_DATASET_ID,
                        partition="train",
                    ),
                    files=[("somepath", 1337)],
                )
            ],
        )

    def test_shuffle_job_factory_one_partition(self) -> None:
        """Tests the functionality of a shuffle_job_factory when provided with a mock backend"""
        partition = "train"
        mock_rows: List[MetadataDatabaseScanRow] = [
            MetadataDatabaseScanRow(partition=partition, row_data_path=f"somepath{i}", row_size=i) for i in range(10)
        ]
        mock_dataset_writer_backend = unittest.mock.MagicMock()
        mock_dataset_writer_backend._metadata_db.scan_sorted.return_value = (row for row in mock_rows)
        job_factory = ShuffleJobFactory(mock_dataset_writer_backend)
        self.assertEqual(
            list(job_factory.build_shuffle_jobs(FAKE_DATASET_ID)),
            [
                ShuffleJob(
                    dataset_partition=DatasetPartition(
                        dataset_id=FAKE_DATASET_ID,
                        partition=partition,
                    ),
                    files=[(row.row_data_path, row.row_size) for row in mock_rows],
                )
            ],
        )

    def test_shuffle_job_factory_one_partition_two_working_sets(self) -> None:
        """Tests the functionality of a shuffle_job_factory when provided with a mock backend"""
        partition = "train"
        worker_max_working_set_size = 5
        mock_rows: List[MetadataDatabaseScanRow] = [
            MetadataDatabaseScanRow(partition=partition, row_data_path=f"somepath{i}", row_size=i) for i in range(10)
        ]
        mock_dataset_writer_backend = unittest.mock.MagicMock()
        mock_dataset_writer_backend._metadata_db.scan_sorted.return_value = (row for row in mock_rows)
        job_factory = ShuffleJobFactory(
            mock_dataset_writer_backend, worker_max_working_set_size=worker_max_working_set_size
        )
        self.assertEqual(
            list(job_factory.build_shuffle_jobs(FAKE_DATASET_ID)),
            [
                ShuffleJob(
                    dataset_partition=DatasetPartition(
                        dataset_id=FAKE_DATASET_ID,
                        partition=partition,
                    ),
                    files=[(row.row_data_path, row.row_size) for row in mock_rows[:worker_max_working_set_size]],
                ),
                ShuffleJob(
                    dataset_partition=DatasetPartition(
                        dataset_id=FAKE_DATASET_ID,
                        partition=partition,
                    ),
                    files=[(row.row_data_path, row.row_size) for row in mock_rows[worker_max_working_set_size:]],
                ),
            ],
        )

    def test_shuffle_job_factory_multi_partition_multi_working_sets(self) -> None:
        """Tests the functionality of a shuffle_job_factory when provided with a mock backend"""
        partitions = ("train", "test")
        num_rows = 10
        num_batches = 2
        worker_max_working_set_size = num_rows // num_batches
        mock_rows: List[MetadataDatabaseScanRow] = [
            MetadataDatabaseScanRow(partition=partition, row_data_path=f"somepath{i}", row_size=i)
            for partition in partitions
            for i in range(10)
        ]
        mock_dataset_writer_backend = unittest.mock.MagicMock()
        mock_dataset_writer_backend._metadata_db.scan_sorted.return_value = (row for row in mock_rows)
        job_factory = ShuffleJobFactory(
            mock_dataset_writer_backend, worker_max_working_set_size=worker_max_working_set_size
        )
        self.assertEqual(
            list(job_factory.build_shuffle_jobs(FAKE_DATASET_ID)),
            [
                ShuffleJob(
                    dataset_partition=DatasetPartition(
                        dataset_id=FAKE_DATASET_ID,
                        partition=partition,
                    ),
                    files=[
                        (row.row_data_path, row.row_size)
                        for row in mock_rows[
                            worker_max_working_set_size
                            * (batch + (partition_index * num_batches)) : worker_max_working_set_size
                            * (batch + (partition_index * num_batches) + 1)
                        ]
                    ],
                )
                for partition_index, partition in enumerate(partitions)
                for batch in range(2)
            ],
        )


class TestShuffleWorker(unittest.TestCase):
    def setUp(self) -> None:
        self.boto3_patcher = unittest.mock.patch("wicker.core.shuffle.boto3")
        self.boto3_mock = self.boto3_patcher.start()
        self.uploaded_column_bytes_files: Dict[Tuple[str, str], bytes] = {}

    def tearDown(self) -> None:
        self.boto3_patcher.stop()
        self.uploaded_column_bytes_files.clear()

    @staticmethod
    def download_fileobj_mock(bucket: str, key: str, bio: IO) -> None:
        bio.write(
            pickle.dumps(
                dataparsing.parse_example(
                    FAKE_EXAMPLE,
                    FAKE_DATASET_SCHEMA,
                )
            )
        )
        bio.seek(0)
        return None

    def test_process_job(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            # The threaded workers each construct their own S3DataStorage from a boto3 client to download
            # file data in parallel.  We mock those out here by mocking out the boto3 client itself.
            self.boto3_mock.session.Session().client().download_fileobj.side_effect = self.download_fileobj_mock

            fake_job = ShuffleJob(
                dataset_partition=DatasetPartition(
                    dataset_id=FAKE_DATASET_ID,
                    partition="test",
                ),
                files=[(f"s3://somebucket/path/{i}", i) for i in range(10)],
            )

            fake_storage = FakeS3DataStorage(tmpdir=tmpdir)
            fake_storage.put_object_s3(
                serialization.dumps(FAKE_DATASET_SCHEMA).encode("utf-8"),
                f"{get_config().aws_s3_config.s3_datasets_path}/{FAKE_DATASET_ID.name}"
                f"/{FAKE_DATASET_ID.version}/avro_schema.json",
            )
            worker = ShuffleWorker(storage=fake_storage)
            shuffle_results = worker.process_job(fake_job)

            self.assertEqual(shuffle_results["timestamp"].to_pylist(), [FAKE_EXAMPLE["timestamp"] for _ in range(10)])
            self.assertEqual(shuffle_results["car_id"].to_pylist(), [FAKE_EXAMPLE["car_id"] for _ in range(10)])
            for location_bytes in shuffle_results["vector"].to_pylist():
                location = ColumnBytesFileLocationV1.from_bytes(location_bytes)
                path = worker.s3_path_factory.get_column_concatenated_bytes_s3path_from_uuid(location.file_id.bytes)
                self.assertTrue(fake_storage.check_exists_s3(path))
                data = fake_storage.fetch_obj_s3(path)[location.byte_offset : location.byte_offset + location.data_size]
                self.assertEqual(VectorCodec(0).decode_object(data), FAKE_EXAMPLE["vector"])



"""This module contains the code required to interact with column files

column files are simple files that are just a simple continuous stream of bytes. They cannot
be read without the corresponding index that is stored in a separate metadata store, which contains the bytes
offsets to access the individual rows within the file.
"""

from __future__ import annotations

import dataclasses
import os
import struct
import tempfile
import uuid
from types import TracebackType
from typing import IO, Any, Dict, List, Literal, Optional, Tuple, Type

from wicker.core.storage import S3DataStorage, S3PathFactory
from wicker.schema import schema, validation


@dataclasses.dataclass(order=True)
class ColumnBytesFileLocationV1:
    """Location serialized as:

    1byte protocol version | 16bytes file ID | 8bytes byte offset | 4bytes size of data in bytes
    """

    file_id: uuid.UUID
    byte_offset: int
    data_size: int

    PROTOCOL_VERSION = 1

    # Packed as (all little-endian):
    #   char(1)
    #   char[](16)
    #   unsigned_long_long(8)
    #   unsigned_int(8)
    STRUCT_PACK_FMT = "<c16sQI"

    def to_bytes(self) -> bytes:
        return struct.pack(
            ColumnBytesFileLocationV1.STRUCT_PACK_FMT,
            ColumnBytesFileLocationV1.PROTOCOL_VERSION.to_bytes(1, "little"),
            self.file_id.bytes,
            self.byte_offset,
            self.data_size,
        )

    @classmethod
    def from_bytes(cls, b: bytes) -> ColumnBytesFileLocationV1:
        protocol_version = int.from_bytes(b[0:1], "little")
        if protocol_version != 1:
            raise ValueError(f"Unable to parse ColumnBytesFileLocation with protocol_version={protocol_version}")
        _, file_id, byte_offset, data_size = struct.unpack(ColumnBytesFileLocationV1.STRUCT_PACK_FMT, b)
        return cls(
            file_id=uuid.UUID(bytes=file_id),
            byte_offset=byte_offset,
            data_size=data_size,
        )


@dataclasses.dataclass
class ColumnBytesFileWriteBuffer:
    file_id: uuid.UUID
    write_count: int
    buffer: IO[bytes]

    def append_to_buffer(self, data: bytes) -> Tuple[int, int]:
        """Appends data to the buffer, returning the start and size of the data

        :param data: data to append
        :return: (start, size) of the data in the buffer
        """
        start_byte_offset = self.buffer.tell()
        bytes_written = self.buffer.write(data)
        self.write_count += 1
        return start_byte_offset, bytes_written


class ColumnBytesFileWriter:
    """A class that writes columns into ColumnBytesFile files, with one file per column"""

    def __init__(
        self,
        storage: S3DataStorage = S3DataStorage(),
        s3_path_factory: S3PathFactory = S3PathFactory(),
        target_file_size: Optional[int] = None,
        target_file_rowgroup_size: Optional[int] = None,
    ) -> None:
        """Create a ColumnBytesFileWriter

        :param storage: storage client to use, defaults to S3DataStorage()
        :param s3_path_factory: path factory to use, defaults to S3PathFactory()
        :param target_file_size: If set, open a new binary file when a column file gets larger than this many bytes.
        :param target_file_rowgroup_size: If set, open a new binary file when a column file has more rows than this
        """
        self.storage = storage
        self.s3_path_factory = s3_path_factory
        # {column_name: (file_id, write_count, <filehandle_to_tmp_file>)}
        self.write_buffers: Dict[str, ColumnBytesFileWriteBuffer] = {}
        self.target_file_size = target_file_size
        self.target_file_rowgroup_size = target_file_rowgroup_size

    def __enter__(self) -> ColumnBytesFileWriter:
        return self

    def __exit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ) -> Literal[False]:
        self.close()
        return False

    def add(self, column_name: str, data: bytes) -> ColumnBytesFileLocationV1:
        """Adds some data into the write buffer

        :param column_name: name of the column to write to
        :param data: data to write into column
        :return: info about location in the binary file
        """
        # Initialize if first element to be added
        if column_name not in self.write_buffers:
            self.write_buffers[column_name] = self._get_new_buffer()

        # Update the io.BytesIO buffer
        write_buffer = self.write_buffers[column_name]
        start_byte_offset, bytes_written = write_buffer.append_to_buffer(data)
        file_size_reached_limit = (
            self.target_file_size is not None and start_byte_offset + bytes_written >= self.target_file_size
        )
        file_num_rows_reached_limit = (
            self.target_file_rowgroup_size is not None and write_buffer.write_count >= self.target_file_rowgroup_size
        )
        if file_size_reached_limit or file_num_rows_reached_limit:
            self._write_column(column_name)

        return ColumnBytesFileLocationV1(
            file_id=write_buffer.file_id,
            byte_offset=start_byte_offset,
            data_size=bytes_written,
        )

    def close(self) -> None:
        """Writes the data in the buffer to S3"""
        for column_name in self.write_buffers:
            self._write_column(column_name)

    def _get_new_buffer(self) -> ColumnBytesFileWriteBuffer:
        return ColumnBytesFileWriteBuffer(
            file_id=uuid.uuid4(),
            write_count=0,
            buffer=tempfile.NamedTemporaryFile(),
        )

    def _write_column(self, column_name: str) -> None:
        columns_root_path = self.s3_path_factory.get_column_concatenated_bytes_files_path()
        write_buffer = self.write_buffers[column_name]
        path = os.path.join(columns_root_path, str(write_buffer.file_id))
        if write_buffer.buffer.tell() > 0:
            write_buffer.buffer.flush()
            write_buffer.buffer.seek(0)
            self.storage.put_file_s3(
                write_buffer.buffer.name,
                path,
            )
        # Get a new UUID so that the next chunk of this column does not overwrite the previous chunks.
        write_buffer.buffer.close()
        self.write_buffers[column_name] = self._get_new_buffer()


class ColumnBytesFileCache:
    """A read-through caching abstraction for accessing ColumnBytesFiles"""

    def __init__(
        self,
        local_cache_path_prefix: str = "/tmp",
        filelock_timeout_seconds: int = -1,
        path_factory: S3PathFactory = S3PathFactory(),
        storage: Optional[S3DataStorage] = None,
    ):
        """Initializes a ColumnBytesFileCache

        :param local_cache_path_prefix: root to path on disk to use as a disk cache, defaults to "/tmp"
        :type local_cache_path_prefix: str, optional
        :param filelock_timeout_seconds: number of seconds after which to timeout on waiting for downloads,
            defaults to -1 which indicates to wait forever
        :type filelock_timeout_seconds: int, optional
        """
        self._s3_storage = storage if storage is not None else S3DataStorage()
        self._root_path = local_cache_path_prefix
        self._filelock_timeout_seconds = filelock_timeout_seconds
        self._columns_root_path = path_factory.get_column_concatenated_bytes_files_path()

    def read(
        self,
        cbf_info: ColumnBytesFileLocationV1,
    ) -> bytes:
        column_concatenated_bytes_file_path = os.path.join(self._columns_root_path, str(cbf_info.file_id))

        local_path = self._s3_storage.fetch_file_s3(
            column_concatenated_bytes_file_path,
            self._root_path,
            timeout_seconds=self._filelock_timeout_seconds,
        )

        with open(local_path, "rb") as f:
            f.seek(cbf_info.byte_offset)
            return f.read(cbf_info.data_size)

    def resolve_pointers(
        self,
        example: validation.AvroRecord,
        schema: schema.DatasetSchema,
    ) -> validation.AvroRecord:
        visitor = ResolvePointersVisitor(example, schema, self)
        return visitor.resolve_pointers()


class ResolvePointersVisitor(schema.DatasetSchemaVisitor[Any]):
    """schema.DatasetSchemaVisitor class that will resolve all heavy pointers in an example by
    downloading the appropriate ColumnBytesFile files and retrieving the appropriate bytes
    """

    def __init__(
        self,
        example: validation.AvroRecord,
        schema: schema.DatasetSchema,
        cbf_cache: ColumnBytesFileCache,
    ):
        self.cbf_cache = cbf_cache

        # Pointers to original data (data should be kept immutable)
        self._schema = schema
        self._example = example

        # Pointers to help keep visitor state during tree traversal
        self._current_data: Any = self._example
        self._current_path: Tuple[str, ...] = tuple()

    def resolve_pointers(self) -> Dict[str, Any]:
        """Resolve all heavy pointers"""
        # Since the original input example is non-None, the loaded example will be non-None also
        example: Dict[str, Any] = self._schema.schema_record._accept_visitor(self)
        return example

    def process_record_field(self, field: schema.RecordField) -> Optional[validation.AvroRecord]:
        """Visit an schema.RecordField schema field"""
        current_data = validation.validate_dict(self._current_data, field.required, self._current_path)
        if current_data is None:
            return current_data

        # Process nested fields by setting up the visitor's state and visiting each node
        processing_path = self._current_path
        processing_example = current_data
        loaded = {}

        # When reading records, the client might restrict the columns to load to a subset of the
        # full columns, so check if the key is actually present in the example being processed
        for nested_field in field.fields:
            if nested_field.name in processing_example:
                self._current_path = processing_path + (nested_field.name,)
                self._current_data = processing_example[nested_field.name]
                loaded[nested_field.name] = nested_field._accept_visitor(self)
        return loaded

    def process_array_field(self, field: schema.ArrayField) -> Optional[List[Any]]:
        current_data = validation.validate_field_type(self._current_data, list, field.required, self._current_path)
        if current_data is None:
            return current_data

        # Process array elements by setting up the visitor's state and visiting each element
        processing_path = self._current_path
        loaded = []

        # Arrays may contain None values if the element field declares that it is not required
        for element_index, element in enumerate(current_data):
            self._current_path = processing_path + (f"elem[{element_index}]",)
            self._current_data = element
            loaded.append(field.element_field._accept_visitor(self))
        return loaded

    def process_int_field(self, field: schema.IntField) -> Any:
        return self._current_data

    def process_long_field(self, field: schema.LongField) -> Any:
        return self._current_data

    def process_string_field(self, field: schema.StringField) -> Any:
        return self._current_data

    def process_bool_field(self, field: schema.BoolField) -> Any:
        return self._current_data

    def process_float_field(self, field: schema.FloatField) -> Any:
        return self._current_data

    def process_double_field(self, field: schema.DoubleField) -> Any:
        return self._current_data

    def process_object_field(self, field: schema.ObjectField) -> Any:
        if not field.is_heavy_pointer:
            return self._current_data
        data = validation.validate_field_type(self._current_data, bytes, field.required, self._current_path)
        if data is None:
            return data
        cbf_info = ColumnBytesFileLocationV1.from_bytes(data)
        return self.cbf_cache.read(cbf_info)



from __future__ import annotations

import abc
import io
import json
import re
from typing import Any, Dict, Generic, List, Optional, Tuple, Type, TypeVar

import numpy as np

from wicker.core.errors import WickerSchemaException
from wicker.schema import codecs

__all__ = [
    "DatasetSchema",
    "IntField",
    "LongField",
    "StringField",
    "BoolField",
    "FloatField",
    "DoubleField",
    "RecordField",
    "ArrayField",
    "ObjectField",
    "NumpyField",
    "BytesField",
]

_T = TypeVar("_T")

PRIMARY_KEYS_TAG = "_primary_keys"


class SchemaField(abc.ABC):
    """Base class for all schema fields"""

    def __init__(
        self,
        name: str,
        description: str = "",
        required: bool = True,
        custom_field_tags: Dict[str, str] = {},
    ):
        """Create a new SchemaField

        :param name: name of the field
        :param description: description of the field, defaults to ""
        :param required: whether this field is required, defaults to True
        :param custom_field_tags: custom field tags to dump in schema, defaults to {}
        """
        # We reserve names starting with _ for internal use.
        if not re.match(r"^[a-zA-Z][0-9a-zA-Z_]*$", name):
            raise ValueError(
                "Schema name must contain only alphanumeric characters or underscores and cannot start with a number or"
                " underscore"
            )

        self.name: str = name
        self.description: str = description
        self.required = required
        self.custom_field_tags: Dict[str, str] = custom_field_tags

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False
        elif self.name != other.name:
            return False
        elif self.description != other.description:
            return False
        elif self.required != other.required:
            return False
        return True

    @property
    def is_heavy_pointer(self) -> bool:
        """Whether this field is a heavy bytes field, in which case we should store a pointer and
        offload the contents to a separate file for more efficient columnar access

        This defaults to False for most fields, but individual SchemaField implementations should
        override this if they serialize to bytes and require offloading into a separate file.
        """
        return False

    @abc.abstractmethod
    def _accept_visitor(self, visitor: DatasetSchemaVisitor[_T]) -> _T:
        """Processes the current schema field with the visitor object"""
        raise NotImplementedError()


class IntField(SchemaField):
    """A field that stores a 32-bit int"""

    def _accept_visitor(self, visitor: DatasetSchemaVisitor[_T]) -> _T:
        """Processes the current schema field with the visitor object"""
        return visitor.process_int_field(self)


class LongField(SchemaField):
    """A field that stores a 64-bit long"""

    def _accept_visitor(self, visitor: DatasetSchemaVisitor[_T]) -> _T:
        """Processes the current schema field with the visitor object"""
        return visitor.process_long_field(self)


class StringField(SchemaField):
    """A field that stores a string"""

    def _accept_visitor(self, visitor: DatasetSchemaVisitor[_T]) -> _T:
        """Processes the current schema field with the visitor object"""
        return visitor.process_string_field(self)


class BoolField(SchemaField):
    """A field that stores a boolean"""

    def _accept_visitor(self, visitor: DatasetSchemaVisitor[_T]) -> _T:
        """Processes the current schema field with the visitor object"""
        return visitor.process_bool_field(self)


class FloatField(SchemaField):
    """A field that stores a 32-bit float"""

    def _accept_visitor(self, visitor: DatasetSchemaVisitor[_T]) -> _T:
        """Processes the current schema field with the visitor object"""
        return visitor.process_float_field(self)


class DoubleField(SchemaField):
    """A field that stores a 64-bit double"""

    def _accept_visitor(self, visitor: DatasetSchemaVisitor[_T]) -> _T:
        """Processes the current schema field with the visitor object"""
        return visitor.process_double_field(self)


class RecordField(SchemaField):
    """A field that contains nested fields"""

    def __init__(
        self,
        name: str,
        fields: List[SchemaField],
        description: str = "",
        required: bool = True,
        custom_field_tags: Dict[str, str] = {},
        top_level: bool = False,
    ):
        """Create a RecordField

        :param name: name of the field
        :param fields: list of nested SchemaFields
        :param description: description of the field, defaults to ""
        :param required: whether this field can be None, defaults to True
        :param custom_field_tags: custom field tags to dump in schema, defaults to {}
        :param top_level: whether this is the top-level field container, defaults to False
        """
        super().__init__(name, description=description, required=required, custom_field_tags=custom_field_tags)

        if not top_level and any([f.is_heavy_pointer for f in fields]):
            raise WickerSchemaException(
                f"Heavy pointer fields cannot be nested in a "
                f"RecordField: {[f.name for f in fields if f.is_heavy_pointer]}"
            )

        self.fields: List[SchemaField] = fields

    def _accept_visitor(self, visitor: DatasetSchemaVisitor[_T]) -> _T:
        """Processes the current schema field with the visitor object"""
        return visitor.process_record_field(self)

    def __eq__(self, other: Any) -> bool:
        return (
            super().__eq__(other)
            and len(self.fields) == len(other.fields)
            and all([self_field == other_field for self_field, other_field in zip(self.fields, other.fields)])
        )


class ArrayField(SchemaField):
    """A field that contains an array of values (that have the same schema type)"""

    def __init__(self, element_field: SchemaField, required: bool = True):
        super().__init__(
            element_field.name,
            description=element_field.description,
            required=required,
            custom_field_tags=element_field.custom_field_tags,
        )
        self.element_field = element_field

    def _accept_visitor(self, visitor: DatasetSchemaVisitor[_T]) -> _T:
        """Processes the current schema field with the visitor object"""
        return visitor.process_array_field(self)

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other) and self.element_field == other.element_field


class ObjectField(SchemaField):
    """A field that contains objects of a specific type."""

    def __init__(
        self,
        name: str,
        codec: codecs.Codec,
        description: str = "",
        required: bool = True,
        is_heavy_pointer: bool = True,
    ) -> None:
        """Create an ObjectField. The ObjectField is parametrized with a Codec that it will use when
        serializing/deserializing data to/from storage. Users are responsible for providing the Codec
        at both write and read time.

        :param name: name of the field
        :param codec: Encoder/decoder to serialize/deserialize the object. See codecs.Codec for more details on
        the codecs.
        """
        if not codec.get_codec_name():
            raise WickerSchemaException("Codec names must be non-empty. Encountered at field={name}")

        custom_field_tags = {
            "_l5ml_metatype": "object",
            "_codec_params": json.dumps(codec.save_codec_to_dict()),
            "_codec_name": codec.get_codec_name(),
        }
        super().__init__(
            name,
            description=description,
            required=required,
            custom_field_tags=custom_field_tags,
        )
        self._is_heavy_pointer = is_heavy_pointer
        self.codec = codec

    @property
    def is_heavy_pointer(self) -> bool:
        """Whether to store this field as a separate file"""
        return self._is_heavy_pointer

    def _accept_visitor(self, visitor: DatasetSchemaVisitor[_T]) -> _T:
        """Processes the current schema field with the visitor object"""
        return visitor.process_object_field(self)

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other) and self.codec == other.codec


class WickerNumpyCodec(codecs.Codec):
    def __init__(self, shape: Optional[Tuple[int, ...]], dtype=str):
        self.shape = shape
        # Validate the dtype is a valid numpy dtype
        try:
            self.dtype = np.dtype(dtype)
        except TypeError:
            raise WickerSchemaException(f"Specified dtype: {dtype} not understood by numpy")

    @staticmethod
    def _codec_name() -> str:
        return "wicker_numpy"

    def save_codec_to_dict(self) -> Dict[str, Any]:
        """If you want to save some parameters of this codec with the dataset
        schema, return the fields here. The returned dictionary should be JSON compatible.
        Note that this is a dataset-level value, not a per example value."""
        return {
            "shape": [d for d in self.shape] if self.shape is not None else None,
            "dtype": str(self.dtype),
        }

    @staticmethod
    def load_codec_from_dict(data: Dict[str, Any]) -> WickerNumpyCodec:
        """Create a new instance of this codec with the given parameters."""
        return WickerNumpyCodec(
            shape=tuple(data["shape"]) if data["shape"] is not None else None,
            dtype=data["dtype"],
        )

    def validate_and_encode_object(self, obj: np.ndarray) -> bytes:
        """Encode the given object into bytes. The function is also responsible for validating the data.
        :param obj: Object to encode
        :return: The encoded bytes for the given object."""
        if obj.dtype != self.dtype:
            raise WickerSchemaException(
                f"Example provided a numpy array with dtype {obj.dtype}, " f"expected {self.dtype}"
            )
        if self.shape is not None:
            if len(obj.shape) != len(self.shape):
                raise WickerSchemaException(
                    f"Example provided a numpy array with shape {obj.shape}, "
                    f"which has a different number of dimensions from expected shape {self.shape}"
                )
            for arr_dim_size, field_dim_size in zip(obj.shape, self.shape):
                if field_dim_size == -1:
                    continue
                if arr_dim_size != field_dim_size:
                    raise WickerSchemaException(
                        f"Example provided a numpy array with shape {obj.shape}, "
                        f"which is incompatible with expected shape {self.shape}"
                    )

        # Serialize array as .npy bytes
        bio = io.BytesIO()
        np.save(bio, obj)
        return bio.getvalue()

    def decode_object(self, data: bytes) -> np.ndarray:
        """Decode an object from the given bytes. This is the opposite of validate_and_encode_object.
        We expect obj == decode_object(validate_and_encode_object(obj))
        :param data: bytes to decode.
        :return: Decoded object."""
        return np.load(io.BytesIO(data))

    def object_type(self) -> Type[Any]:
        """Return the expected type of the objects handled by this codec.
        This method can be overriden to match more specific classes."""
        return np.ndarray


class NumpyField(ObjectField):
    """An ObjectField that uses a Codec for encoding Numpy arrays"""

    def __init__(
        self,
        name: str,
        shape: Optional[Tuple[int, ...]],
        dtype: str,
        description: str = "",
        required: bool = True,
        is_heavy_pointer: bool = True,
    ) -> None:
        """Create a NumpyField

        :param name: name of the field
        :param shape: shape of the numpy array that we expect, or None to indicate that all shapes are acceptable,
            `-1` denotes that a given dimension can have any size.
        :param dtype: dtype of the numpy array that we expect,
        """
        super().__init__(
            name=name,
            codec=WickerNumpyCodec(shape=shape, dtype=dtype),
            description=description,
            required=required,
            is_heavy_pointer=is_heavy_pointer,
        )


class WickerNoopBytesCodec(codecs.Codec):
    @staticmethod
    def _codec_name() -> str:
        return "wicker_bytes"

    def save_codec_to_dict(self) -> Dict[str, Any]:
        """If you want to save some parameters of this codec with the dataset
        schema, return the fields here. The returned dictionary should be JSON compatible.
        Note that this is a dataset-level value, not a per example value."""
        return {}

    @staticmethod
    def load_codec_from_dict(data: Dict[str, Any]) -> WickerNoopBytesCodec:
        """Create a new instance of this codec with the given parameters."""
        return WickerNoopBytesCodec()

    def validate_and_encode_object(self, obj: bytes) -> bytes:
        """Encode the given object into bytes. The function is also responsible for validating the data.
        :param obj: Object to encode
        :return: The encoded bytes for the given object."""
        return obj

    def decode_object(self, data: bytes) -> bytes:
        """Decode an object from the given bytes. This is the opposite of validate_and_encode_object.
        We expect obj == decode_object(validate_and_encode_object(obj))
        :param data: bytes to decode.
        :return: Decoded object."""
        return data

    def object_type(self) -> Type[Any]:
        """Return the expected type of the objects handled by this codec.
        This method can be overriden to match more specific classes."""
        return bytes


class BytesField(ObjectField):
    """An ObjectField that uses a no-op Codec for encoding pure bytes"""

    def __init__(
        self,
        name: str,
        description: str = "",
        required: bool = True,
        is_heavy_pointer: bool = True,
    ) -> None:
        """Create a BytesField

        :param name: name of the field
        """
        super().__init__(
            name=name,
            codec=WickerNoopBytesCodec(),
            description=description,
            required=required,
            is_heavy_pointer=is_heavy_pointer,
        )


class DatasetSchema:
    """A schema definition that serializes into an Avro-compatible schema"""

    def __init__(
        self,
        fields: List[SchemaField],
        primary_keys: List[str],
        allow_empty_primary_keys: bool = False,
    ) -> None:
        """Create a new dataset schema.
        :param field: List of fields at the top level of the structure.
        :param primary_keys: List of field names to use to order the data. Keys must be listed in order of precedence,
            and must be of type int, long, str or bool.
            The keys are used to uniquely identify examples, they are also used when joining multiple datasets.
        :param allow_empty_primary_keys: Internal use. Do not change.
        """
        self.schema_record = RecordField(
            name="fields",
            fields=fields,
            custom_field_tags={
                PRIMARY_KEYS_TAG: json.dumps(primary_keys),
            },
            top_level=True,
        )
        self._columns = {f.name: f for f in fields}
        self.primary_keys = primary_keys
        self._validate_schema(allow_empty_primary_keys)

    def _validate_schema(self, allow_empty_primary_keys: bool) -> None:
        # We want all new datasets to have primary keys. But the allow_empty_primary_keys flag allows us to construct
        # DatasetSchemas from existing files which did not have this properly specified. Clean me AVSW-78939.
        if not allow_empty_primary_keys and not self.primary_keys:
            raise WickerSchemaException("The primary_keys attribute can not be empty.")
        for key in self.primary_keys:
            if key not in self._columns:
                raise WickerSchemaException(f"Primary key '{key}' not found in the schema's fields.")
            field = self._columns[key]
            if not field.required:
                raise WickerSchemaException(
                    f"All primary key fields must have the 'required' tag, but '{key}' doesn't have it."
                )
            # In particular, we don't accept floats/double as keys since inconsistent rounding could be an issue.
            supported_types = [IntField, LongField, StringField, BoolField]
            if not any(isinstance(field, t) for t in supported_types):
                raise WickerSchemaException(
                    f"'{key}' cannot be a primary key because it is not of type int, long, string or bool."
                )

    def get_all_column_names(self) -> List[str]:
        """Returns all column names"""
        return list(self._columns.keys())

    def get_column(self, column_name: str) -> Optional[SchemaField]:
        """Get a column by name"""
        return self._columns.get(column_name)

    def get_pointer_columns(self) -> List[str]:
        """Get the list of columms that contain heavy pointer data."""
        return [col_name for col_name, field in self._columns.items() if field.is_heavy_pointer]

    def get_non_pointer_columns(self) -> List[str]:
        """Get the list of columms that don't contain heavy pointer data. These are usually small metadata columns."""
        return [col_name for col_name, field in self._columns.items() if not field.is_heavy_pointer]

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, DatasetSchema)
            and self.schema_record == other.schema_record
            and self.primary_keys == other.primary_keys
        )


class DatasetSchemaVisitor(abc.ABC, Generic[_T]):
    """Class that operates over SchemaField objects using the Visitor pattern"""

    @abc.abstractmethod
    def process_int_field(self, field: IntField) -> _T:
        pass

    @abc.abstractmethod
    def process_long_field(self, field: LongField) -> _T:
        pass

    @abc.abstractmethod
    def process_string_field(self, field: StringField) -> _T:
        pass

    @abc.abstractmethod
    def process_bool_field(self, field: BoolField) -> _T:
        pass

    @abc.abstractmethod
    def process_float_field(self, field: FloatField) -> _T:
        pass

    @abc.abstractmethod
    def process_double_field(self, field: DoubleField) -> _T:
        pass

    @abc.abstractmethod
    def process_object_field(self, field: ObjectField) -> _T:
        pass

    @abc.abstractmethod
    def process_record_field(self, field: RecordField) -> _T:
        pass

    @abc.abstractmethod
    def process_array_field(self, field: ArrayField) -> _T:
        pass



from typing import Any, Dict, Optional, Tuple, Type, TypeVar

from wicker.core.errors import WickerSchemaException

T = TypeVar("T")


# Unfortunately, mypy has weak support for recursive typing so we resort to Any here, but Any
# should really be an alias for: Union[AvroRecord, str, bool, int, float, bytes]
AvroRecord = Dict[str, Any]


def validate_field_type(val: Any, type_: Type[T], required: bool, current_path: Tuple[str, ...]) -> Optional[T]:
    """Validates the type of a field

    :param val: value to validate
    :type val: Any
    :param type_: type to validate against
    :type type_: Type[T]
    :param required: whether or not the value is required (to be non-None)
    :type required: bool
    :param current_path: current parsing path
    :type current_path: Tuple[str, ...]
    :raises WickerSchemaException: when parsing error occurs
    :return: val, but validated to be of type T
    :rtype: T
    """
    if val is None:
        if not required:
            return val
        raise WickerSchemaException(
            f"Error at path {'.'.join(current_path)}: Example provided a None value for required field"
        )
    elif not isinstance(val, type_):
        raise WickerSchemaException(
            f"Error at path {'.'.join(current_path)}: Example provided a {type(val)} value, expected {type_}"
        )
    return val


def validate_dict(val: Any, required: bool, current_path: Tuple[str, ...]) -> Optional[Dict[str, Any]]:
    """Validates a dictionary

    :param val: incoming value to validate as a dictionary
    :type val: Any
    :param required: whether or not the value is required (to be non-None)
    :type required: bool
    :param current_path: current parsing path
    :type current_path: Tuple[str, ...]
    :return: parsed dictionary
    :rtype: dict
    """
    if val is None:
        if not required:
            return val
        raise WickerSchemaException(
            f"Error at path {'.'.join(current_path)}: Example provided a None value for required field"
        )
    # PyArrow returns record fields as lists of (k, v) tuples
    elif isinstance(val, list):
        try:
            parsed_val = dict(val)
        except ValueError:
            raise WickerSchemaException(f"Error at path {'.'.join(current_path)}: Unable to convert list to dict")
    elif isinstance(val, dict):
        parsed_val = val
    else:
        raise WickerSchemaException(f"Error at path {'.'.join(current_path)}: Unable to convert object to dict")

    return validate_field_type(parsed_val, dict, required, current_path)



import copy
import json
import unittest
from typing import Any, Dict

import avro.schema  # type: ignore

from wicker import schema
from wicker.core.errors import WickerSchemaException
from wicker.schema import dataloading, dataparsing, serialization
from wicker.schema.schema import PRIMARY_KEYS_TAG
from wicker.testing.codecs import Vector, VectorCodec

TEST_SCHEMA = schema.DatasetSchema(
    fields=[
        schema.IntField("label", description="Label of the example"),
        schema.RecordField(
            "lidar_point_cloud",
            fields=[
                schema.RecordField(
                    "lidar_metadata",
                    fields=[
                        schema.StringField(
                            "lidar_model",
                            description="Model of lidar used to generate data",
                        ),
                        schema.DoubleField(
                            "lidar_calibration_error",
                            description="Some lidar calibration metric",
                            required=False,
                        ),
                    ],
                    description="Some metadata about the lidar",
                ),
            ],
            description="Lidar point cloud data",
        ),
        schema.LongField("timestamp_ns", description="Some timestamp field in ns"),
        schema.FloatField("ego_speed", description="Absolute speed of ego"),
        schema.BoolField("qc", description="A quality control field", required=False),
        schema.RecordField(
            "extra_metadata",
            description="Extra metadata",
            fields=[
                schema.IntField("meta_1", description="Metadata 1"),
                schema.IntField("meta_2", description="Metadata 2", required=False),
            ],
            required=False,
        ),
        schema.ArrayField(
            schema.StringField("array_stringfield", description="some array"),
        ),
    ],
    primary_keys=["timestamp_ns"],
)
TEST_EXAMPLE_REQUIRED: Dict[str, Any] = {
    "label": 0,
    "lidar_point_cloud": {
        "lidar_metadata": {
            "lidar_model": "harambe",
        },
    },
    "timestamp_ns": 1337,
    "ego_speed": 1337.1337,
    "array_stringfield": ["foo", "bar", "baz"],
}
# When we load the example all the keys for any unset non-required fields will are added
TEST_EXAMPLE_LOAD_REQUIRED = copy.deepcopy(TEST_EXAMPLE_REQUIRED)
TEST_EXAMPLE_LOAD_REQUIRED.update(
    {
        "lidar_point_cloud": {
            "lidar_metadata": {
                "lidar_model": "harambe",
                "lidar_calibration_error": None,
            },
        },
        "qc": None,
        "extra_metadata": None,
    }
)
# Example with everything field set
TEST_EXAMPLE_FULL = copy.deepcopy(TEST_EXAMPLE_REQUIRED)
TEST_EXAMPLE_FULL.update(
    {
        "lidar_point_cloud": {
            "lidar_metadata": {
                "lidar_model": "harambe",
                "lidar_calibration_error": 1.337,
            },
        },
        "qc": True,
        "extra_metadata": {
            "meta_1": 1,
            "meta_2": 2,
        },
    }
)

TEST_SERIALIZED_JSON_V2 = {
    "_description": "",
    "_json_version": 2,
    PRIMARY_KEYS_TAG: '["timestamp_ns"]',
    "fields": [
        {"_description": "Label of the example", "name": "label", "type": "int"},
        {
            "_description": "Lidar point cloud data",
            "name": "lidar_point_cloud",
            "type": {
                "fields": [
                    {
                        "_description": "Some metadata about the lidar",
                        "name": "lidar_metadata",
                        "type": {
                            "fields": [
                                {
                                    "_description": "Model of lidar used to generate data",
                                    "name": "lidar_model",
                                    "type": "string",
                                },
                                {
                                    "_description": "Some lidar calibration metric",
                                    "name": "lidar_calibration_error",
                                    "type": ["null", "double"],
                                },
                            ],
                            "name": "lidar_metadata",
                            "type": "record",
                        },
                    }
                ],
                "name": "lidar_point_cloud",
                "type": "record",
            },
        },
        {"_description": "Some timestamp field in ns", "name": "timestamp_ns", "type": "long"},
        {"_description": "Absolute speed of ego", "name": "ego_speed", "type": "float"},
        {"_description": "A quality control field", "name": "qc", "type": ["null", "boolean"]},
        {
            "_description": "Extra metadata",
            "name": "extra_metadata",
            "type": [
                "null",
                {
                    "fields": [
                        {"_description": "Metadata 1", "name": "meta_1", "type": "int"},
                        {"_description": "Metadata 2", "name": "meta_2", "type": ["null", "int"]},
                    ],
                    "name": "extra_metadata",
                    "type": "record",
                },
            ],
        },
        {
            "name": "array_stringfield",
            "type": {"_description": "some array", "items": "string", "name": "array_stringfield", "type": "array"},
        },
    ],
    "name": "fields",
    "type": "record",
}


class TestSchemaParseExample(unittest.TestCase):
    def setUp(self) -> None:
        self.maxDiff = None

    def test_parse_full_example(self) -> None:
        parsed_example = dataparsing.parse_example(TEST_EXAMPLE_FULL, TEST_SCHEMA)
        self.assertEqual(parsed_example, TEST_EXAMPLE_FULL)

    def test_parse_required_fields(self) -> None:
        parsed_example = dataparsing.parse_example(TEST_EXAMPLE_REQUIRED, TEST_SCHEMA)
        self.assertEqual(parsed_example, TEST_EXAMPLE_LOAD_REQUIRED)

    def test_fail_required(self) -> None:
        example = copy.deepcopy(TEST_EXAMPLE_REQUIRED)
        del example["label"]
        with self.assertRaises(WickerSchemaException) as e:
            dataparsing.parse_example(example, TEST_SCHEMA)
        self.assertIn("Example missing keys", str(e.exception))

        example = copy.deepcopy(TEST_EXAMPLE_REQUIRED)
        del example["lidar_point_cloud"]
        with self.assertRaises(WickerSchemaException) as e:
            dataparsing.parse_example(example, TEST_SCHEMA)
        self.assertIn("Example missing keys", str(e.exception))

    def test_fail_type_int(self) -> None:
        example = copy.deepcopy(TEST_EXAMPLE_FULL)
        example["label"] = "SHOULD_BE_INT"
        with self.assertRaises(WickerSchemaException) as e:
            dataparsing.parse_example(example, TEST_SCHEMA)
        self.assertIn("Error at path label", str(e.exception))

    def test_fail_type_long(self) -> None:
        example = copy.deepcopy(TEST_EXAMPLE_FULL)
        example["timestamp_ns"] = "SHOULD_BE_LONG"
        with self.assertRaises(WickerSchemaException) as e:
            dataparsing.parse_example(example, TEST_SCHEMA)
        self.assertIn("Error at path timestamp_ns", str(e.exception))

    def test_fail_type_bool(self) -> None:
        example = copy.deepcopy(TEST_EXAMPLE_FULL)
        example["qc"] = "SHOULD_BE_BOOL"
        with self.assertRaises(WickerSchemaException) as e:
            dataparsing.parse_example(example, TEST_SCHEMA)
        self.assertIn("Error at path qc", str(e.exception))

    def test_fail_type_float(self) -> None:
        example = copy.deepcopy(TEST_EXAMPLE_FULL)
        example["ego_speed"] = "SHOULD_BE_FLOAT"
        with self.assertRaises(WickerSchemaException) as e:
            dataparsing.parse_example(example, TEST_SCHEMA)
        self.assertIn("Error at path ego_speed", str(e.exception))

    def test_fail_type_double(self) -> None:
        example: Dict[str, Any] = copy.deepcopy(TEST_EXAMPLE_FULL)
        example["lidar_point_cloud"]["lidar_metadata"]["lidar_calibration_error"] = "SHOULD_BE_DOUBLE"
        with self.assertRaises(WickerSchemaException) as e:
            dataparsing.parse_example(example, TEST_SCHEMA)
        self.assertIn(
            "Error at path lidar_point_cloud.lidar_metadata.lidar_calibration_error",
            str(e.exception),
        )

    def test_fail_type_string(self) -> None:
        example: Dict[str, Any] = copy.deepcopy(TEST_EXAMPLE_FULL)
        example["lidar_point_cloud"]["lidar_metadata"]["lidar_model"] = 1337
        with self.assertRaises(WickerSchemaException) as e:
            dataparsing.parse_example(example, TEST_SCHEMA)
        self.assertIn(
            "Error at path lidar_point_cloud.lidar_metadata.lidar_model",
            str(e.exception),
        )

    def test_fail_type_record(self) -> None:
        example = copy.deepcopy(TEST_EXAMPLE_FULL)
        example["lidar_point_cloud"] = "SHOULD_BE_DICT"
        with self.assertRaises(WickerSchemaException) as e:
            dataparsing.parse_example(example, TEST_SCHEMA)
        self.assertIn("Error at path lidar_point_cloud", str(e.exception))

    def test_fail_keys_record(self) -> None:
        example = copy.deepcopy(TEST_EXAMPLE_FULL)
        del example["label"]
        with self.assertRaises(WickerSchemaException) as e:
            dataparsing.parse_example(example, TEST_SCHEMA)
        self.assertIn("Error at path :", str(e.exception))

    def test_fail_type_array(self) -> None:
        example = copy.deepcopy(TEST_EXAMPLE_FULL)
        example["array_stringfield"] = "foo"
        with self.assertRaises(WickerSchemaException) as e:
            dataparsing.parse_example(example, TEST_SCHEMA)
        self.assertIn("Error at path array_stringfield:", str(e.exception))

    def test_fail_element_type_array(self) -> None:
        example = copy.deepcopy(TEST_EXAMPLE_FULL)
        example["array_stringfield"] = [1, 2, 3]
        with self.assertRaises(WickerSchemaException) as e:
            dataparsing.parse_example(example, TEST_SCHEMA)
        self.assertIn("Error at path array_stringfield.elem[0]:", str(e.exception))


class TestSchemaValidation(unittest.TestCase):
    def test_schema_no_primary_keys(self) -> None:
        with self.assertRaises(WickerSchemaException) as e:
            schema.DatasetSchema(fields=[], primary_keys=[])
        self.assertIn("The primary_keys attribute can not be empty", str(e.exception))

    def test_schema_invalid_primary_keys(self) -> None:
        with self.assertRaises(WickerSchemaException) as e:
            schema.DatasetSchema(fields=[], primary_keys=["unknown_key"])
        self.assertIn("'unknown_key' not found", str(e.exception))

    def test_schema_required_primary_keys(self) -> None:
        schema.DatasetSchema(fields=[schema.StringField("car_id")], primary_keys=["car_id"])
        with self.assertRaises(WickerSchemaException) as e:
            schema.DatasetSchema(
                fields=[schema.StringField("car_id", required=False)],
                primary_keys=["car_id"],
            )
        self.assertIn("must have the 'required' tag, but 'car_id' doesn't", str(e.exception))

    def test_schema_invalid_primary_keys_type(self) -> None:
        bad_fields = [
            schema.FloatField("float_key"),
            schema.DoubleField("double_key"),
            schema.RecordField("record_key", fields=[]),
        ]
        for f in bad_fields:
            with self.assertRaises(WickerSchemaException, msg=f"field.name={f.name}") as e:
                schema.DatasetSchema(fields=[f], primary_keys=[f.name])
            self.assertIn(f"'{f.name}' cannot be a primary key", str(e.exception), msg=f"field.name={f.name}")


class TestSchemaSerialization(unittest.TestCase):
    def setUp(self) -> None:
        self.maxDiff = None

    def test_schema_name_error_dashes(self) -> None:
        with self.assertRaises(ValueError):
            schema.StringField(name="name-with-dashes")

    def test_schema_name_error_start_with_number(self) -> None:
        with self.assertRaises(ValueError):
            schema.StringField(name="0foo")
        with self.assertRaises(ValueError):
            schema.StringField(name="0")

    def test_schema_name_single_char(self) -> None:
        schema.StringField(name="q")

    def test_serialize_to_str(self) -> None:
        serialized = serialization.dumps(TEST_SCHEMA)
        avro.schema.parse(serialized)
        self.assertEqual(json.loads(serialized), TEST_SERIALIZED_JSON_V2)

    def test_serialize_to_str_pretty(self) -> None:
        serialized = serialization.dumps(TEST_SCHEMA, pretty=True)
        avro.schema.parse(serialized)
        self.assertEqual(serialized, json.dumps(TEST_SERIALIZED_JSON_V2, sort_keys=True, indent=4))

    def test_loads(self) -> None:
        self.assertEqual(
            TEST_SCHEMA,
            serialization.loads(json.dumps(TEST_SERIALIZED_JSON_V2)),
        )

    def test_loads_bad_type(self) -> None:
        with self.assertRaises(WickerSchemaException):
            serialized_bad_type = copy.deepcopy(TEST_SERIALIZED_JSON_V2)
            serialized_bad_type["fields"][0]["type"] = "BAD_TYPE_123"  # type: ignore
            serialization.loads(json.dumps(serialized_bad_type))

    def test_loads_bad_type_nullable(self) -> None:
        with self.assertRaises(WickerSchemaException):
            serialized_bad_type = copy.deepcopy(TEST_SERIALIZED_JSON_V2)
            serialized_bad_type["fields"][0]["type"] = ["null", "BAD_TYPE_123"]  # type: ignore
            serialization.loads(json.dumps(serialized_bad_type))


class TestSchemaLoading(unittest.TestCase):
    def test_load(self) -> None:
        loaded_example = dataloading.load_example(TEST_EXAMPLE_FULL, TEST_SCHEMA)

        # Assert that values are semantically equal
        self.assertEqual(TEST_EXAMPLE_FULL, loaded_example)

        # Assert that the IDs of the values are also equal (no additional copies of the data were created)
        self.assertEqual(id(TEST_EXAMPLE_FULL["label"]), id(loaded_example["label"]))
        self.assertEqual(
            id(TEST_EXAMPLE_FULL["lidar_point_cloud"]["lidar_metadata"]["lidar_model"]),
            id(loaded_example["lidar_point_cloud"]["lidar_metadata"]["lidar_model"]),
        )
        self.assertEqual(
            id(TEST_EXAMPLE_FULL["lidar_point_cloud"]["lidar_metadata"]["lidar_calibration_error"]),
            id(loaded_example["lidar_point_cloud"]["lidar_metadata"]["lidar_calibration_error"]),
        )
        self.assertEqual(id(TEST_EXAMPLE_FULL["timestamp_ns"]), id(loaded_example["timestamp_ns"]))
        self.assertEqual(id(TEST_EXAMPLE_FULL["ego_speed"]), id(loaded_example["ego_speed"]))
        self.assertEqual(id(TEST_EXAMPLE_FULL["qc"]), id(loaded_example["qc"]))

        # Assert that the IDs of dictionaries and lists are not equal (data is not modified in-place)
        self.assertNotEqual(id(TEST_EXAMPLE_FULL), id(loaded_example))
        self.assertNotEqual(
            id(TEST_EXAMPLE_FULL["lidar_point_cloud"]),
            id(loaded_example["lidar_point_cloud"]),
        )
        self.assertNotEqual(
            id(TEST_EXAMPLE_FULL["lidar_point_cloud"]["lidar_metadata"]),
            id(loaded_example["lidar_point_cloud"]["lidar_metadata"]),
        )
        self.assertNotEqual(id(TEST_EXAMPLE_FULL["extra_metadata"]), id(loaded_example["extra_metadata"]))
        self.assertNotEqual(id(TEST_EXAMPLE_FULL["array_stringfield"]), id(loaded_example["array_stringfield"]))

    def test_load_columns_required(self) -> None:
        subset_example = copy.deepcopy(TEST_EXAMPLE_REQUIRED)
        del subset_example["label"]
        loaded_example = dataloading.load_example(subset_example, TEST_SCHEMA)

        # Assert that values are semantically equal
        self.assertEqual(subset_example, loaded_example)

    def test_load_extra_keys_ignored(self) -> None:
        extra_keys_example = copy.deepcopy(TEST_EXAMPLE_FULL)
        extra_keys_example["extra_key_foo"] = 1
        loaded_example = dataloading.load_example(extra_keys_example, TEST_SCHEMA)

        # Assert that values are semantically equal
        self.assertEqual(TEST_EXAMPLE_FULL, loaded_example)

    def test_load_record_from_list_kv_tuples(self) -> None:
        example = copy.deepcopy(TEST_EXAMPLE_FULL)
        example["lidar_point_cloud"]["lidar_metadata"] = [
            (k, v) for k, v in example["lidar_point_cloud"]["lidar_metadata"].items()
        ]
        example["lidar_point_cloud"] = [(k, v) for k, v in example["lidar_point_cloud"].items()]
        parsed_example = dataloading.load_example(example, TEST_SCHEMA)
        self.assertEqual(parsed_example, TEST_EXAMPLE_FULL)


def test_schema(testcase: unittest.TestCase, schema_to_test: schema.DatasetSchema) -> None:
    serialized = serialization.dumps(schema_to_test, pretty=True)
    loaded = serialization.loads(serialized)
    loaded_serialized = serialization.dumps(schema_to_test, pretty=True)
    testcase.assertEqual(loaded, schema_to_test, msg=f"{serialized} vs {loaded_serialized}")
    testcase.assertTrue(json.loads(serialized))
    testcase.assertTrue(avro.schema.parse(serialized))


class TestObjectSchemas(unittest.TestCase):
    OBJECT_FIELD = schema.ObjectField("encoded_vector", VectorCodec(compression_method=12), required=False)
    SCHEMA = schema.DatasetSchema(
        fields=[OBJECT_FIELD, schema.StringField("sample_id")],
        primary_keys=["sample_id"],
    )
    EXAMPLE = {"sample_id": "sample000", "encoded_vector": Vector([1, 2, 3, 4])}
    EXAMPLE_BAD_TYPE = {"sample_id": "sample000", "encoded_vector": [1, 2, 3, 4]}
    EXAMPLE_NONE = {"sample_id": "sample000"}

    def test_serialization(self) -> None:
        serialized = serialization.dumps(TestObjectSchemas.SCHEMA, pretty=True)
        loaded = serialization.loads(serialized)
        loaded_serialized = serialization.dumps(TestObjectSchemas.SCHEMA, pretty=True)
        self.assertEqual(loaded, TestObjectSchemas.SCHEMA, msg=f"{serialized} vs {loaded_serialized}")
        self.assertTrue(json.loads(serialized))
        self.assertTrue(avro.schema.parse(serialized))

    def test_deserialization_unknown_codec(self) -> None:
        # Test the case where we want to deserialize a schema and we don't have the necessary codec.
        serialized = serialization.dumps(TestObjectSchemas.SCHEMA, pretty=True)
        serialized = serialized.replace("VectorCodec", "UnknownCodec")
        # By default we should fail
        with self.assertRaises(WickerSchemaException):
            loaded = serialization.loads(serialized)
        # But with the treat_objects_as_bytes=True we should be able to do it.
        loaded = serialization.loads(serialized, treat_objects_as_bytes=True)
        serialized2 = serialization.dumps(loaded, pretty=True)
        # Make sure that if we reserialize the schema that got loaded with treat_objects_as_bytes=True, we get the
        # same thing as the original schema.
        self.assertEqual(serialized, serialized2)

    def test_good_example1(self) -> None:
        # Check parsing and loading
        parsed_example = dataparsing.parse_example(TestObjectSchemas.EXAMPLE, TestObjectSchemas.SCHEMA)
        self.assertTrue(isinstance(parsed_example["encoded_vector"], bytes))
        loaded_schema = serialization.loads(serialization.dumps(TestObjectSchemas.SCHEMA, pretty=True))
        loaded_example = dataloading.load_example(parsed_example, loaded_schema)
        self.assertEqual(loaded_example, TestObjectSchemas.EXAMPLE)
        assert isinstance(loaded_schema.schema_record.fields[0], schema.ObjectField)  # Make mypy happy
        assert isinstance(loaded_schema.schema_record.fields[0].codec, VectorCodec)
        self.assertEqual(loaded_schema.schema_record.fields[0].codec.compression_method, 12)

    def test_example_none(self) -> None:
        parsed_example = dataparsing.parse_example(TestObjectSchemas.EXAMPLE_NONE, TestObjectSchemas.SCHEMA)
        loaded_example = dataloading.load_example(parsed_example, TestObjectSchemas.SCHEMA)
        self.assertEqual(loaded_example, TestObjectSchemas.EXAMPLE_NONE)

    def test_bad_validation(self) -> None:
        with self.assertRaises(WickerSchemaException):
            dataparsing.parse_example(TestObjectSchemas.EXAMPLE_BAD_TYPE, TestObjectSchemas.SCHEMA)

    def test_loads_bad_l5ml_metatype(self) -> None:
        with self.assertRaises(WickerSchemaException) as err:
            serialized_bad_type = json.loads(serialization.dumps(TestObjectSchemas.SCHEMA))
            serialized_bad_type["fields"][0]["_l5ml_metatype"] = "BAD_TYPE_123"  # type: ignore
            serialization.loads(json.dumps(serialized_bad_type))
        self.assertIn("Unhandled _l5ml_metatype for avro bytes type: BAD_TYPE_123", str(err.exception))



import io
import unittest

import numpy as np

from wicker.core.errors import WickerSchemaException
from wicker.schema.schema import WickerNumpyCodec

EYE_ARR = np.eye(4)
eye_arr_bio = io.BytesIO()
np.save(eye_arr_bio, EYE_ARR)
EYE_ARR_BYTES = eye_arr_bio.getvalue()


class TestNumpyCodec(unittest.TestCase):
    def test_codec_none_shape(self) -> None:
        codec = WickerNumpyCodec(shape=None, dtype="float64")
        self.assertEqual(codec.get_codec_name(), "wicker_numpy")
        self.assertEqual(codec.object_type(), np.ndarray)
        self.assertEqual(
            codec.save_codec_to_dict(),
            {
                "dtype": "float64",
                "shape": None,
            },
        )
        self.assertEqual(WickerNumpyCodec.load_codec_from_dict(codec.save_codec_to_dict()), codec)
        self.assertEqual(codec.validate_and_encode_object(EYE_ARR), EYE_ARR_BYTES)
        np.testing.assert_equal(codec.decode_object(EYE_ARR_BYTES), EYE_ARR)

    def test_codec_unbounded_dim_shape(self) -> None:
        codec = WickerNumpyCodec(shape=(-1, -1), dtype="float64")
        self.assertEqual(codec.get_codec_name(), "wicker_numpy")
        self.assertEqual(codec.object_type(), np.ndarray)
        self.assertEqual(
            codec.save_codec_to_dict(),
            {
                "dtype": "float64",
                "shape": [-1, -1],
            },
        )
        self.assertEqual(WickerNumpyCodec.load_codec_from_dict(codec.save_codec_to_dict()), codec)
        self.assertEqual(codec.validate_and_encode_object(EYE_ARR), EYE_ARR_BYTES)
        np.testing.assert_equal(codec.decode_object(EYE_ARR_BYTES), EYE_ARR)

        # Should raise when provided with bad shapes with too few/many dimensions
        with self.assertRaises(WickerSchemaException):
            codec.validate_and_encode_object(np.ones((10,)))
        with self.assertRaises(WickerSchemaException):
            codec.validate_and_encode_object(np.ones((10, 10, 10)))

    def test_codec_fixed_shape(self) -> None:
        codec = WickerNumpyCodec(shape=(4, 4), dtype="float64")
        self.assertEqual(codec.get_codec_name(), "wicker_numpy")
        self.assertEqual(codec.object_type(), np.ndarray)
        self.assertEqual(
            codec.save_codec_to_dict(),
            {
                "dtype": "float64",
                "shape": [4, 4],
            },
        )
        self.assertEqual(WickerNumpyCodec.load_codec_from_dict(codec.save_codec_to_dict()), codec)
        self.assertEqual(codec.validate_and_encode_object(EYE_ARR), EYE_ARR_BYTES)
        np.testing.assert_equal(codec.decode_object(EYE_ARR_BYTES), EYE_ARR)

        # Should raise when provided with bad shapes with too few/many/wrong dimensions
        with self.assertRaises(WickerSchemaException):
            codec.validate_and_encode_object(np.ones((10,)))
        with self.assertRaises(WickerSchemaException):
            codec.validate_and_encode_object(np.ones((10, 10, 10)))
        with self.assertRaises(WickerSchemaException):
            codec.validate_and_encode_object(np.ones((5, 4)))

    def test_codec_bad_dtype(self) -> None:
        with self.assertRaises(WickerSchemaException):
            WickerNumpyCodec(shape=(4, 4), dtype="SOME_BAD_DTYPE")



import abc
from typing import Any, Dict, Iterable, List, Optional, Tuple

import pyarrow as pa
import pyarrow.compute as pc

from wicker import schema as schema_module
from wicker.core.column_files import ColumnBytesFileWriter
from wicker.core.definitions import DatasetID
from wicker.core.shuffle import save_index
from wicker.core.storage import S3DataStorage, S3PathFactory
from wicker.schema import dataparsing, serialization

PARTITION_SIZE = 256
MAX_COL_FILE_NUMROW = 50  # TODO(isaak-willett): Magic number, we should derive this based on row size

UnparsedExample = Dict[str, Any]
ParsedExample = Dict[str, Any]
PointerParsedExample = Dict[str, Any]


class AbstractDataPersistor(abc.ABC):
    """
    Abstract class for persisting data onto a user defined cloud or local instance.

    Only s3 is supported right now but plan to support other data stores
    (BigQuery, Azure, Postgres)
    """

    def __init__(
        self,
        s3_storage: S3DataStorage = S3DataStorage(),
        s3_path_factory: S3PathFactory = S3PathFactory(),
    ) -> None:
        """
        Init a Persister

        :param s3_storage: The storage abstraction for S3
        :type s3_storage: S3DataStore
        :param s3_path_factory: The path factory for generating s3 paths
                                based on dataset name and version
        :type s3_path_factory: S3PathFactory
        """
        super().__init__()
        self.s3_storage = s3_storage
        self.s3_path_factory = s3_path_factory

    @abc.abstractmethod
    def persist_wicker_dataset(
        self,
        dataset_name: str,
        dataset_version: str,
        dataset_schema: schema_module.DatasetSchema,
        dataset: Any,
    ) -> Optional[Dict[str, int]]:
        """
        Persist a user specified dataset defined by name, version, schema, and data.

        :param dataset_name: Name of the dataset
        :type dataset_name: str
        :param dataset_version: Version of the dataset
        :type: dataset_version: str
        :param dataset_schema: Schema of the dataset
        :type dataset_schema: wicker.schema.schema.DatasetSchema
        :param dataset: Data of the dataset
        :type dataset: User defined
        """
        raise NotImplementedError("Method, persist_wicker_dataset, needs to be implemented in inhertiance class.")

    @staticmethod
    def parse_row(data_row: UnparsedExample, schema: schema_module.DatasetSchema) -> ParsedExample:
        """
        Parse a row to test for validation errors.

        :param data_row: Data row to be parsed
        :type data_row: UnparsedExample
        :return: parsed row containing the correct types associated with schema
        :rtype: ParsedExample
        """
        return dataparsing.parse_example(data_row, schema)

        # Write data to Column Byte Files

    @staticmethod
    def persist_wicker_partition(
        spark_partition_iter: Iterable[Tuple[str, ParsedExample]],
        schema: schema_module.DatasetSchema,
        s3_storage: S3DataStorage,
        s3_path_factory: S3PathFactory,
        target_max_column_file_numrows: int = 50,
    ) -> Iterable[Tuple[str, PointerParsedExample]]:
        """Persists a Spark partition of examples with parsed bytes into S3Storage as ColumnBytesFiles,
        returning a new Spark partition of examples with heavy-pointers and metadata only.
        :param spark_partition_iter: Spark partition of `(partition_str, example)`, where `example`
            is a dictionary of parsed bytes that needs to be uploaded to S3
        :param target_max_column_file_numrows: Maximum number of rows in column files. Defaults to 50.
        :return: a Generator of `(partition_str, example)`, where `example` is a dictionary with heavy-pointers
            that point to ColumnBytesFiles in S3 in place of the parsed bytes
        """
        column_bytes_file_writers: Dict[str, ColumnBytesFileWriter] = {}
        heavy_pointer_columns = schema.get_pointer_columns()
        metadata_columns = schema.get_non_pointer_columns()

        for partition, example in spark_partition_iter:
            # Create ColumnBytesFileWriter lazily as required, for each partition
            if partition not in column_bytes_file_writers:
                column_bytes_file_writers[partition] = ColumnBytesFileWriter(
                    s3_storage,
                    s3_path_factory,
                    target_file_rowgroup_size=target_max_column_file_numrows,
                )

            # Write to ColumnBytesFileWriter and return only metadata + heavy-pointers
            parquet_metadata: Dict[str, Any] = {col: example[col] for col in metadata_columns}
            for col in heavy_pointer_columns:
                loc = column_bytes_file_writers[partition].add(col, example[col])
                parquet_metadata[col] = loc.to_bytes()
            yield partition, parquet_metadata

        # Flush all writers when finished
        for partition in column_bytes_file_writers:
            column_bytes_file_writers[partition].close()

    @staticmethod
    def save_partition_tbl(
        partition_table_tuple: Tuple[str, pa.Table],
        dataset_name: str,
        dataset_version: str,
        s3_storage: S3DataStorage,
        s3_path_factory: S3PathFactory,
    ) -> Tuple[str, int]:
        """
        Save a partition table to s3 under the dataset name and version.

        :param partition_table_tuple: Tuple of partition id and pyarrow table to save
        :type partition_table_tuple: Tuple[str, pyarrow.Table]
        :return: A tuple containing the paritiion id and the num of saved rows
        :rtype: Tuple[str, int]
        """
        partition, pa_tbl = partition_table_tuple
        save_index(
            dataset_name,
            dataset_version,
            {partition: pa_tbl},
            s3_storage=s3_storage,
            s3_path_factory=s3_path_factory,
        )
        return (partition, pa_tbl.num_rows)


def persist_wicker_dataset(
    dataset_name: str,
    dataset_version: str,
    dataset_schema: schema_module.DatasetSchema,
    dataset: Any,
    s3_storage: S3DataStorage = S3DataStorage(),
    s3_path_factory: S3PathFactory = S3PathFactory(),
) -> Optional[Dict[str, int]]:
    """
    Persist wicker dataset public facing api function, for api consistency.
    :param dataset_name: name of dataset persisted
    :type dataset_name: str
    :param dataset_version: version of dataset persisted
    :type dataset_version: str
    :param dataset_schema: schema of dataset to be persisted
    :type dataset_schema: DatasetSchema
    :param rdd: rdd of data to persist
    :type rdd: RDD
    :param s3_storage: s3 storage abstraction
    :type s3_storage: S3DataStorage
    :param s3_path_factory: s3 path abstraction
    :type s3_path_factory: S3PathFactory
    """
    return BasicPersistor(s3_storage, s3_path_factory).persist_wicker_dataset(
        dataset_name, dataset_version, dataset_schema, dataset
    )


class BasicPersistor(AbstractDataPersistor):
    """
    Basic persistor class that persists wicker data on s3 in a non sorted manner.

    We will move to supporting other features like shuffling, other data engines, etc...
    """

    def __init__(
        self, s3_storage: S3DataStorage = S3DataStorage(), s3_path_factory: S3PathFactory = S3PathFactory()
    ) -> None:
        super().__init__(s3_storage, s3_path_factory)

    def persist_wicker_dataset(
        self, dataset_name: str, dataset_version: str, dataset_schema: schema_module.DatasetSchema, dataset: Any
    ) -> Optional[Dict[str, int]]:
        """
        Persist a user defined dataset, pushing data to s3 in a basic manner

        :param dataset_name: Name of the dataset
        :type dataset_name: str
        :param dataset_version: Version of the dataset
        :type: dataset_version: str
        :param dataset_schema: Schema of the dataset
        :type dataset_schema: wicker.schema.schema.DatasetSchema
        :param dataset: Data of the dataset
        :type dataset: User defined
        """
        # what needs to be done within this function
        # 1. Check if the variables are set
        # check if variables have been set ie: not None
        if (
            not isinstance(dataset_name, str)
            or not isinstance(dataset_version, str)
            or not isinstance(dataset_schema, schema_module.DatasetSchema)
        ):
            raise ValueError("Current dataset variables not all set, set all to proper not None values")

        # 2. Put the schema up on s3
        schema_path = self.s3_path_factory.get_dataset_schema_path(
            DatasetID(name=dataset_name, version=dataset_version)
        )
        self.s3_storage.put_object_s3(serialization.dumps(dataset_schema).encode("utf-8"), schema_path)

        # 3. Validate the rows and ensure data is well formed, sort while doing
        dataset_0 = [(row[0], self.parse_row(row[1], dataset_schema)) for row in dataset]

        # 4. Sort the dataset if not sorted
        sorted_dataset_0 = sorted(dataset_0, key=lambda tup: tup[0])

        # 5. Partition the dataset into K partitions
        partitions = []

        def divide_chunks(list_to_divide):
            # looping till length l
            for i in range(0, len(list_to_divide), PARTITION_SIZE):
                partitions.append(list_to_divide[i : i + PARTITION_SIZE])

        divide_chunks(sorted_dataset_0)

        # 6. Persist the partitions to S3
        for partition in partitions:
            # build a persistence iterator for each parition
            iterator = self.persist_wicker_partition(
                partition, dataset_schema, self.s3_storage, self.s3_path_factory, MAX_COL_FILE_NUMROW
            )
            # make sure all yields get called
            list(iterator)

        # 7. Create the parition table, need to combine keys in a way we can form table
        # split into k dicts where k is partition number and the data is a list of values
        # for each key for all the dicts in the partition
        merged_dicts: Dict[str, Dict[str, List[Any]]] = {}
        for partition_key, row in sorted_dataset_0:
            current_dict: Dict[str, List[Any]] = merged_dicts.get(partition_key, {})
            for col in row.keys():
                if col in current_dict:
                    current_dict[col].append(row[col])
                else:
                    current_dict[col] = [row[col]]
            merged_dicts[partition_key] = current_dict
        # convert each of the dicts to a pyarrow table in the same way SparkPersistor
        # converts, needed to ensure parity between the two
        arrow_dict = {}
        for partition_key, data_dict in merged_dicts.items():
            data_table = pa.Table.from_pydict(data_dict)
            arrow_dict[partition_key] = pc.take(
                pa.Table.from_pydict(data_dict),
                pc.sort_indices(data_table, sort_keys=[(pk, "ascending") for pk in dataset_schema.primary_keys]),
            )

        # 8. Persist the partition table to s3
        written_dict = {}
        for partition_key, pa_table in arrow_dict.items():
            self.save_partition_tbl(
                (partition_key, pa_table), dataset_name, dataset_version, self.s3_storage, self.s3_path_factory
            )
            written_dict[partition_key] = pa_table.num_rows

        return written_dict



from __future__ import annotations

import dataclasses
import enum
import re
from typing import Any, Dict, TypeVar

# flake8 linting struggles with the module name `schema` and DatasetDefinition dataclass field name `schema`
from wicker import schema  # noqa: 401

Example = TypeVar("Example")
ExampleMetadata = Dict[str, Any]


DATASET_ID_REGEX = re.compile(r"^(?P<name>[0-9a-zA-Z_]+)/(?P<version>[0-9]+\.[0-9]+\.[0-9]+)$")


@dataclasses.dataclass(frozen=True)
class DatasetID:
    """Representation of the unique identifier of a dataset

    `name` should be alphanumeric and contain no spaces, only underscores
    `version` should be a semantic version (e.g. 1.0.0)
    """

    name: str
    version: str

    @classmethod
    def from_str(cls, s: str) -> DatasetID:
        """Parses a DatasetID from a string"""
        match = DATASET_ID_REGEX.match(s)
        if not match:
            raise ValueError(f"{s} is not a valid DatasetID")
        return cls(name=match["name"], version=match["version"])

    def __str__(self) -> str:
        """Helper function to return the representation of the dataset version as a path-like string"""
        return f"{self.name}/{self.version}"

    @staticmethod
    def validate_dataset_id(name: str, version: str) -> None:
        """Validates the name and version of a dataset"""
        if not re.match(r"^[0-9a-zA-Z_]+$", name):
            raise ValueError(
                f"Provided dataset name {name} must be alphanumeric and contain no spaces, only underscores"
            )
        if not re.match(r"^[0-9]+\.[0-9]+\.[0-9]+$", version):
            raise ValueError(
                f"Provided dataset version {version} should be a semantic version without any prefixes/suffixes"
            )

    def __post_init__(self) -> None:
        DatasetID.validate_dataset_id(self.name, self.version)


@dataclasses.dataclass(frozen=True)
class DatasetDefinition:
    """Representation of the definition of a dataset (immutable once dataset is added)

    `name` should be alphanumeric and contain no spaces, only underscores
    `version` should be a semantic version (e.g. 1.0.0)
    """

    dataset_id: DatasetID
    schema: schema.DatasetSchema

    @property
    def identifier(self) -> DatasetID:
        return DatasetID(name=self.dataset_id.name, version=self.dataset_id.version)

    def __post_init__(self) -> None:
        DatasetID.validate_dataset_id(self.dataset_id.name, self.dataset_id.version)


@dataclasses.dataclass(frozen=True)
class DatasetPartition:
    """Representation of the definition of a partition within dataset

    The partition here is meant to represent the common train/val/test splits of a dataset, but
    could also represent partitions for other use cases.

    `partition` should be alphanumeric and contain no spaces, only underscores
    """

    dataset_id: DatasetID
    partition: str

    def __str__(self) -> str:
        """Helper function to return the representation of the partition as a path-like string"""
        return f"{self.dataset_id.name}/{self.dataset_id.version}/{self.partition}"


class DatasetState(enum.Enum):
    """Representation of the state of a dataset"""

    STAGED = "PENDING"
    COMMITTED = "COMMITTED"



from typing import Any, Dict, List, Optional, Tuple, TypeVar

from wicker.schema import schema, validation

T = TypeVar("T")


def load_example(example: validation.AvroRecord, schema: schema.DatasetSchema) -> Dict[str, Any]:
    """Loads an example according to the provided schema, converting data based on their column
    information into their corresponding in-memory representations (e.g. numpy arrays, torch tensors)

    The returned dictionary will be in the same shape as defined in the schema, with keys
    corresponding to the schema names, and values validated/transformed against the schema's fields.

    :param example: example to load
    :type example: validation.AvroRecord
    :param schema: schema to load against
    :type schema: schema.DatasetSchema
    :return: loaded example
    :rtype: Dict[str, Any]
    """
    load_example_visitor = LoadExampleVisitor(example, schema)
    return load_example_visitor.load_example()


class LoadExampleVisitor(schema.DatasetSchemaVisitor[Any]):
    """schema.DatasetSchemaVisitor class that will validate and load an Example
    in accordance to a provided schema.DatasetSchema
    """

    def __init__(self, example: validation.AvroRecord, schema: schema.DatasetSchema):
        # Pointers to original data (data should be kept immutable)
        self._schema = schema
        self._example = example

        # Pointers to help keep visitor state during tree traversal
        self._current_data: Any = self._example
        self._current_path: Tuple[str, ...] = tuple()

    def load_example(self) -> Dict[str, Any]:
        """Loads an example from its Avro format into its in-memory representations"""
        # Since the original input example is non-None, the loaded example will be non-None also
        example: Dict[str, Any] = self._schema.schema_record._accept_visitor(self)
        return example

    def process_record_field(self, field: schema.RecordField) -> Optional[validation.AvroRecord]:
        """Visit an schema.RecordField schema field"""
        current_data = validation.validate_dict(self._current_data, field.required, self._current_path)
        if current_data is None:
            return current_data

        # Process nested fields by setting up the visitor's state and visiting each node
        processing_path = self._current_path
        processing_example = current_data
        loaded = {}

        # When reading records, the client might restrict the columns to load to a subset of the
        # full columns, so check if the key is actually present in the example being processed
        for nested_field in field.fields:
            if nested_field.name in processing_example:
                self._current_path = processing_path + (nested_field.name,)
                self._current_data = processing_example[nested_field.name]
                loaded[nested_field.name] = nested_field._accept_visitor(self)
        return loaded

    def process_int_field(self, field: schema.IntField) -> Optional[int]:
        return validation.validate_field_type(self._current_data, int, field.required, self._current_path)

    def process_long_field(self, field: schema.LongField) -> Optional[int]:
        return validation.validate_field_type(self._current_data, int, field.required, self._current_path)

    def process_string_field(self, field: schema.StringField) -> Optional[str]:
        return validation.validate_field_type(self._current_data, str, field.required, self._current_path)

    def process_bool_field(self, field: schema.BoolField) -> Optional[bool]:
        return validation.validate_field_type(self._current_data, bool, field.required, self._current_path)

    def process_float_field(self, field: schema.FloatField) -> Optional[float]:
        return validation.validate_field_type(self._current_data, float, field.required, self._current_path)

    def process_double_field(self, field: schema.DoubleField) -> Optional[float]:
        return validation.validate_field_type(self._current_data, float, field.required, self._current_path)

    def process_object_field(self, field: schema.ObjectField) -> Optional[Any]:
        data = validation.validate_field_type(self._current_data, bytes, field.required, self._current_path)
        if data is None:
            return data
        return field.codec.decode_object(data)

    def process_array_field(self, field: schema.ArrayField) -> Optional[List[Any]]:
        current_data = validation.validate_field_type(self._current_data, list, field.required, self._current_path)
        if current_data is None:
            return current_data

        # Process array elements by setting up the visitor's state and visiting each element
        processing_path = self._current_path
        loaded = []

        # Arrays may contain None values if the element field declares that it is not required
        for element_index, element in enumerate(current_data):
            self._current_path = processing_path + (f"elem[{element_index}]",)
            self._current_data = element
            loaded.append(field.element_field._accept_visitor(self))
        return loaded



import os
import shutil
from typing import Any, Dict

from wicker.core.storage import S3DataStorage


class FakeS3DataStorage(S3DataStorage):
    def __init__(self, tmpdir: str = "/tmp") -> None:
        self._tmpdir = tmpdir

    def __getstate__(self) -> Dict[Any, Any]:
        return {"tmpdir": self._tmpdir}

    def __setstate__(self, state: Dict[Any, Any]) -> None:
        self._tmpdir = state["tmpdir"]
        return

    def _get_local_path(self, path: str) -> str:
        return os.path.join(self._tmpdir, path.replace("s3://", ""))

    def check_exists_s3(self, input_path: str) -> bool:
        return os.path.exists(self._get_local_path(input_path))

    def fetch_obj_s3(self, input_path: str) -> bytes:
        if not self.check_exists_s3(input_path):
            raise KeyError(f"File {input_path} not found in the fake s3 storage.")
        with open(self._get_local_path(input_path), "rb") as f:
            return f.read()

    def fetch_file_s3(self, input_path: str, local_prefix: str, timeout_seconds: int = 120) -> str:
        if not self.check_exists_s3(input_path):
            raise KeyError(f"File {input_path} not found in the fake s3 storage.")
        bucket, key = self.bucket_key_from_s3_path(input_path)
        dest_path = os.path.join(local_prefix, key)
        os.makedirs(os.path.dirname(dest_path), exist_ok=True)
        if not os.path.isfile(dest_path):
            shutil.copy2(self._get_local_path(input_path), dest_path)
        return dest_path

    def put_object_s3(self, object_bytes: bytes, s3_path: str) -> None:
        full_tmp_path = self._get_local_path(s3_path)
        os.makedirs(os.path.dirname(full_tmp_path), exist_ok=True)
        with open(full_tmp_path, "wb") as f:
            f.write(object_bytes)

    def put_file_s3(self, local_path: str, s3_path: str) -> None:
        full_tmp_path = self._get_local_path(s3_path)
        os.makedirs(os.path.dirname(full_tmp_path), exist_ok=True)
        shutil.copy2(local_path, full_tmp_path)



import copy
import os
import random
import uuid
from typing import Any, Dict, List, Tuple

import pyarrow.parquet as papq
import pytest

from wicker import schema
from wicker.core.config import get_config
from wicker.core.persistance import BasicPersistor
from wicker.core.storage import S3PathFactory
from wicker.schema.schema import DatasetSchema
from wicker.testing.storage import FakeS3DataStorage

DATASET_NAME = "dataset"
DATASET_VERSION = "0.0.1"
SCHEMA = schema.DatasetSchema(
    primary_keys=["bar", "foo"],
    fields=[
        schema.IntField("foo"),
        schema.StringField("bar"),
        schema.BytesField("bytescol"),
    ],
)
EXAMPLES = [
    (
        "train" if i % 2 == 0 else "test",
        {
            "foo": random.randint(0, 10000),
            "bar": str(uuid.uuid4()),
            "bytescol": b"0",
        },
    )
    for i in range(10000)
]
# Examples with a duplicated key
EXAMPLES_DUPES = copy.deepcopy(EXAMPLES)


@pytest.fixture
def mock_basic_persistor(request, tmpdir) -> Tuple[BasicPersistor, str]:
    storage = request.param.get("storage", FakeS3DataStorage(tmpdir=tmpdir))
    path_factory = request.param.get("path_factory", S3PathFactory())
    return BasicPersistor(storage, path_factory), tmpdir


def assert_written_correctness(tmpdir: str) -> None:
    """Asserts that all files are written as expected by the L5MLDatastore"""
    # Check that files are correctly written locally by Spark/Parquet with a _SUCCESS marker file
    prefix = get_config().aws_s3_config.s3_datasets_path.replace("s3://", "")
    assert DATASET_NAME in os.listdir(os.path.join(tmpdir, prefix))
    assert DATASET_VERSION in os.listdir(os.path.join(tmpdir, prefix, DATASET_NAME))
    for partition in ["train", "test"]:
        print(os.listdir(os.path.join(tmpdir, prefix)))
        columns_path = os.path.join(tmpdir, prefix, "__COLUMN_CONCATENATED_FILES__")
        all_read_bytes = b""
        for filename in os.listdir(columns_path):
            concatenated_bytes_filepath = os.path.join(columns_path, filename)
            with open(concatenated_bytes_filepath, "rb") as bytescol_file:
                all_read_bytes += bytescol_file.read()
        assert all_read_bytes == b"0" * 10000

        # Load parquet file and assert ordering of primary_key
        assert f"{partition}.parquet" in os.listdir(os.path.join(tmpdir, prefix, DATASET_NAME, DATASET_VERSION))
        tbl = papq.read_table(os.path.join(tmpdir, prefix, DATASET_NAME, DATASET_VERSION, f"{partition}.parquet"))
        foobar = [(barval.as_py(), fooval.as_py()) for fooval, barval in zip(tbl["foo"], tbl["bar"])]
        assert foobar == sorted(foobar)


@pytest.mark.parametrize(
    "mock_basic_persistor, dataset_name, dataset_version, dataset_schema, dataset",
    [({}, DATASET_NAME, DATASET_VERSION, SCHEMA, copy.deepcopy(EXAMPLES_DUPES))],
    indirect=["mock_basic_persistor"],
)
def test_basic_persistor(
    mock_basic_persistor: Tuple[BasicPersistor, str],
    dataset_name: str,
    dataset_version: str,
    dataset_schema: DatasetSchema,
    dataset: List[Tuple[str, Dict[str, Any]]],
):
    """
    Test if the basic persistor can persist data in the format we have established.

    Ensure we read the right file locations, the right amount of bytes,
    and the ordering is correct.
    """
    # create the mock basic persistor
    mock_basic_persistor_obj, tempdir = mock_basic_persistor
    # persist the dataset
    mock_basic_persistor_obj.persist_wicker_dataset(dataset_name, dataset_version, dataset_schema, dataset)
    # assert the dataset is correctly written
    assert_written_correctness(tempdir)



from __future__ import annotations

import abc
from typing import Any, Dict, Type


class Codec(abc.ABC):
    """Base class for all object decoder/encoders.

    Defining Codec classes allows user to use arbitrary types in their Wicker Examples. Subclasses of Codec provide
    functionality to serialize/deserialize data to/from underlying storage, and can be used to abstract functionality
    such as compression, validation of object fields and provide adaptor functionality to other libraries (e.g. NumPy,
    PIL, torch).

    Wicker provides some Codec implementations by default for NumPy, but users can choose to define their own Codecs
    and use them during data-writing and data-reading.
    """

    # Maps from codec_name to codec class.
    codec_registry: Dict[str, Type[Codec]] = {}

    def __init_subclass__(cls: Type[Codec], **kwargs: Any) -> None:
        """Automatically register subclasses of Codec."""
        if cls._codec_name() in Codec.codec_registry:
            raise KeyError(f"Codec '{cls._codec_name()}' was already defined.")
        if cls._codec_name():
            Codec.codec_registry[cls._codec_name()] = cls

    @staticmethod
    @abc.abstractmethod
    def _codec_name() -> str:
        """Needs to be overriden to return a globally unique name for the codec."""
        pass

    def get_codec_name(self) -> str:
        """Accessor for _codec_name. In general, derived classes should not touch this, and should just implement
        _codec_name. This accessor is used internally to support generic schema serialization/deserialization.
        """
        return self._codec_name()

    def save_codec_to_dict(self) -> Dict[str, Any]:
        """If you want to save some parameters of this codec with the dataset
        schema, return the fields here. The returned dictionary should be JSON compatible.
        Note that this is a dataset-level value, not a per example value."""
        return {}

    @staticmethod
    @abc.abstractmethod
    def load_codec_from_dict(data: Dict[str, Any]) -> Codec:
        """Create a new instance of this codec with the given parameters."""
        pass

    @abc.abstractmethod
    def validate_and_encode_object(self, obj: Any) -> bytes:
        """Encode the given object into bytes. The function is also responsible for validating the data.
        :param obj: Object to encode
        :return: The encoded bytes for the given object."""
        pass

    @abc.abstractmethod
    def decode_object(self, data: bytes) -> Any:
        """Decode an object from the given bytes. This is the opposite of validate_and_encode_object.
        We expect obj == decode_object(validate_and_encode_object(obj))
        :param data: bytes to decode.
        :return: Decoded object."""
        pass

    def object_type(self) -> Type[Any]:
        """Return the expected type of the objects handled by this codec.
        This method can be overriden to match more specific classes."""
        return object

    def __eq__(self, other: Any) -> bool:
        return (
            super().__eq__(other)
            and self.get_codec_name() == other.get_codec_name()
            and self.save_codec_to_dict() == other.save_codec_to_dict()
        )



from __future__ import annotations

import contextlib
import fcntl
import os
import signal
from typing import Any, Iterator, Optional


@contextlib.contextmanager
def _timeout(seconds: int, lockfile: str) -> Iterator[None]:
    """Context manager for triggering a timeout via SIGALRM if a blocking call
    does not return in `seconds` number of seconds.

    Usage:
    >>> with _timeout(5, "somefile"):
    >>>     some_long_running_function()

    In the above example, if some_long_running_function takes more than 5 seconds, a
    TimeoutError will be raised from the `with _timeout(5, "somefile")` context.
    """
    if seconds == -1.0:
        # No timeout specified, we execute without registering a SIGALRM
        yield
    else:

        def timeout_handler(signum: Any, frame: Any) -> None:
            """Handles receiving a SIGALRM by throwing a TimeoutError with a useful
            error message
            """
            raise TimeoutError(f"{lockfile} could not be acquired after {seconds}s")

        # Register `timeout_handler` as the handler for any SIGALRMs that are raised
        # We store the original SIGALRM handler if available to restore it afterwards
        original_handler = signal.signal(signal.SIGALRM, timeout_handler)

        try:
            # Raise a SIGALRM in `seconds` number of seconds, and then yield to code
            # in the context manager's code-block.
            signal.alarm(seconds)
            yield
        # Make sure to always restore the original SIGALRM handler
        finally:
            signal.alarm(0)
            signal.signal(signal.SIGALRM, original_handler)


class SimpleUnixFileLock:
    """Simple blocking lock class that uses files to achieve system-wide locking

    1. If timeout is not specified, waits forever by default
    2. Does not support recursive locking and WILL break if called recursively
    """

    def __init__(self, lock_file: str, timeout_seconds: int = -1):
        # The path to the lock file.
        self._lock_file = lock_file

        # The file descriptor for the *_lock_file* as it is returned by the
        # os.open() function.
        # This file lock is only NOT None, if the object currently holds the
        # lock.
        self._lock_file_fd: Optional[int] = None

        # The default timeout value.
        self.timeout_seconds = timeout_seconds

        return None

    def __enter__(self) -> SimpleUnixFileLock:
        open_mode = os.O_RDWR | os.O_CREAT | os.O_TRUNC
        fd = os.open(self._lock_file, open_mode)
        try:
            with _timeout(self.timeout_seconds, self._lock_file):
                fcntl.lockf(fd, fcntl.LOCK_EX)
        except (IOError, OSError, TimeoutError):
            os.close(fd)
            raise
        else:
            self._lock_file_fd = fd
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        assert self._lock_file_fd is not None, "self._lock_file_fd not set by __enter__"
        fd = self._lock_file_fd
        self._lock_file_fd = None
        fcntl.flock(fd, fcntl.LOCK_UN)
        os.close(fd)
        return None



"""This module defines how to configure Wicker from the user environment
"""

from __future__ import annotations

import dataclasses
import json
import os
from typing import Any, Dict


@dataclasses.dataclass(frozen=True)
class WickerWandBConfig:
    wandb_base_url: str
    wandb_api_key: str

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> WickerWandBConfig:
        # only load them if they exist, otherwise leave out
        return cls(
            wandb_api_key=data.get("wandb_api_key", None),
            wandb_base_url=data.get("wandb_base_url", None),
        )


@dataclasses.dataclass(frozen=True)
class WickerAwsS3Config:
    s3_datasets_path: str
    region: str

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> WickerAwsS3Config:
        return cls(
            s3_datasets_path=data["s3_datasets_path"],
            region=data["region"],
        )


@dataclasses.dataclass(frozen=True)
class WickerConfig:
    raw: Dict[str, Any]
    aws_s3_config: WickerAwsS3Config
    wandb_config: WickerWandBConfig

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> WickerConfig:
        return cls(
            raw=data,
            aws_s3_config=WickerAwsS3Config.from_json(data["aws_s3_config"]),
            wandb_config=WickerWandBConfig.from_json(data.get("wandb_config", {})),
        )


def get_config() -> WickerConfig:
    """Retrieves the Wicker config for the current process"""

    wicker_config_path = os.getenv("WICKER_CONFIG_PATH", os.path.expanduser("~/wickerconfig.json"))
    with open(wicker_config_path, "r") as f:
        config = WickerConfig.from_json(json.load(f))
    return config



import os
import tempfile
import unittest
from contextlib import contextmanager
from typing import Iterator, Tuple

import numpy as np
import pyarrow as pa  # type: ignore
import pyarrow.fs as pafs  # type: ignore
import pyarrow.parquet as papq  # type: ignore

from wicker.core.column_files import ColumnBytesFileWriter
from wicker.core.datasets import S3Dataset
from wicker.core.definitions import DatasetID, DatasetPartition
from wicker.core.storage import S3PathFactory
from wicker.schema import schema, serialization
from wicker.testing.storage import FakeS3DataStorage

FAKE_NAME = "dataset_name"
FAKE_VERSION = "0.0.1"
FAKE_PARTITION = "train"
FAKE_DATASET_ID = DatasetID(name=FAKE_NAME, version=FAKE_VERSION)
FAKE_DATASET_PARTITION = DatasetPartition(FAKE_DATASET_ID, partition=FAKE_PARTITION)

FAKE_SHAPE = (4, 4)
FAKE_DTYPE = "float64"
FAKE_NUMPY_CODEC = schema.WickerNumpyCodec(FAKE_SHAPE, FAKE_DTYPE)
FAKE_SCHEMA = schema.DatasetSchema(
    primary_keys=["foo"],
    fields=[
        schema.StringField("foo"),
        schema.NumpyField("np_arr", shape=FAKE_SHAPE, dtype=FAKE_DTYPE),
    ],
)
FAKE_DATA = [{"foo": f"bar{i}", "np_arr": np.eye(4)} for i in range(1000)]


@contextmanager
def cwd(path):
    """Changes the current working directory, and returns to the previous directory afterwards"""
    oldpwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpwd)


class TestS3Dataset(unittest.TestCase):
    @contextmanager
    def _setup_storage(self) -> Iterator[Tuple[FakeS3DataStorage, S3PathFactory, str]]:
        """Context manager that sets up a local directory to mimic S3 storage for a committed dataset,
        and returns a tuple of (S3DataStorage, S3PathFactory, tmpdir_path) for the caller to use as
        fixtures in their tests.
        """
        with tempfile.TemporaryDirectory() as tmpdir, cwd(tmpdir):
            fake_s3_storage = FakeS3DataStorage(tmpdir=tmpdir)
            fake_s3_path_factory = S3PathFactory()
            with ColumnBytesFileWriter(
                storage=fake_s3_storage,
                s3_path_factory=fake_s3_path_factory,
                target_file_rowgroup_size=10,
            ) as writer:
                locs = [
                    writer.add("np_arr", FAKE_NUMPY_CODEC.validate_and_encode_object(data["np_arr"]))  # type: ignore
                    for data in FAKE_DATA
                ]
            arrow_metadata_table = pa.Table.from_pydict(
                {
                    "foo": [data["foo"] for data in FAKE_DATA],
                    "np_arr": [loc.to_bytes() for loc in locs],
                }
            )
            metadata_table_path = os.path.join(
                tmpdir,
                fake_s3_path_factory.get_dataset_partition_path(FAKE_DATASET_PARTITION, s3_prefix=False),
                "part-1.parquet",
            )
            os.makedirs(os.path.dirname(metadata_table_path), exist_ok=True)
            papq.write_table(
                arrow_metadata_table,
                metadata_table_path,
            )
            fake_s3_storage.put_object_s3(
                serialization.dumps(FAKE_SCHEMA).encode("utf-8"),
                fake_s3_path_factory.get_dataset_schema_path(FAKE_DATASET_ID),
            )
            yield fake_s3_storage, fake_s3_path_factory, tmpdir

    def test_dataset(self) -> None:
        with self._setup_storage() as (fake_s3_storage, fake_s3_path_factory, tmpdir):
            ds = S3Dataset(
                FAKE_NAME,
                FAKE_VERSION,
                FAKE_PARTITION,
                local_cache_path_prefix=tmpdir,
                columns_to_load=None,
                storage=fake_s3_storage,
                s3_path_factory=None,
                pa_filesystem=pafs.LocalFileSystem(),
            )

            for i in range(len(FAKE_DATA)):
                retrieved = ds[i]
                reference = FAKE_DATA[i]
                self.assertEqual(retrieved["foo"], reference["foo"])
                np.testing.assert_array_equal(retrieved["np_arr"], reference["np_arr"])



import collections
import dataclasses
import json
import tempfile
from typing import Dict, List, Type, cast

try:
    import flytekit  # type: ignore
except ImportError:
    raise RuntimeError(
        "flytekit is not detected in current environment, install Wicker with extra arguments:"
        " `pip install wicker[flyte]`"
    )
import pyarrow as pa  # type: ignore
from flytekit.extend import TypeEngine, TypeTransformer  # type: ignore

from wicker.core.definitions import DatasetID, DatasetPartition
from wicker.core.shuffle import ShuffleJob, ShuffleJobFactory, ShuffleWorker, save_index
from wicker.core.storage import S3DataStorage, S3PathFactory
from wicker.core.writer import DatasetWriterBackend
from wicker.plugins.dynamodb import DynamodbMetadataDatabase

###
# Custom type definitions to clean up passing of data between Flyte tasks
###


class ShuffleJobTransformer(TypeTransformer[ShuffleJob]):
    _TYPE_INFO = flytekit.BlobType(
        format="binary",
        dimensionality=flytekit.BlobType.BlobDimensionality.SINGLE,
    )

    def __init__(self):
        super(ShuffleJobTransformer, self).__init__(name="shufflejob-transform", t=ShuffleJob)

    def get_literal_type(self, t: Type[ShuffleJob]) -> flytekit.LiteralType:
        """
        This is useful to tell the Flytekit type system that ``ShuffleJob`` actually refers to what corresponding type
        In this example, we say its of format binary (do not try to introspect) and there are more than one files in it
        """
        return flytekit.LiteralType(blob=self._TYPE_INFO)

    @staticmethod
    def _shuffle_jobs_to_bytes(job: ShuffleJob) -> bytes:
        return json.dumps(
            {
                "dataset_partition": {
                    "dataset_id": {
                        "name": job.dataset_partition.dataset_id.name,
                        "version": job.dataset_partition.dataset_id.version,
                    },
                    "partition": job.dataset_partition.partition,
                },
                "files": job.files,
            }
        ).encode("utf-8")

    @staticmethod
    def _shuffle_jobs_from_bytes(b: bytes) -> ShuffleJob:
        data = json.loads(b.decode("utf-8"))
        return ShuffleJob(
            dataset_partition=DatasetPartition(
                dataset_id=DatasetID(
                    name=data["dataset_partition"]["dataset_id"]["name"],
                    version=data["dataset_partition"]["dataset_id"]["version"],
                ),
                partition=data["dataset_partition"]["partition"],
            ),
            files=[(path, size) for path, size in data["files"]],
        )

    def to_literal(
        self,
        ctx: flytekit.FlyteContext,
        python_val: ShuffleJob,
        python_type: Type[ShuffleJob],
        expected: flytekit.LiteralType,
    ) -> flytekit.Literal:
        """
        This method is used to convert from given python type object ``MyDataset`` to the Literal representation
        """
        # Step 1: lets upload all the data into a remote place recommended by Flyte
        remote_path = ctx.file_access.get_random_remote_path()
        with tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.write(self._shuffle_jobs_to_bytes(python_val))
            tmpfile.flush()
            tmpfile.seek(0)
            ctx.file_access.upload(tmpfile.name, remote_path)
        # Step 2: lets return a pointer to this remote_path in the form of a literal
        return flytekit.Literal(
            scalar=flytekit.Scalar(
                blob=flytekit.Blob(uri=remote_path, metadata=flytekit.BlobMetadata(type=self._TYPE_INFO))
            )
        )

    def to_python_value(
        self, ctx: flytekit.FlyteContext, lv: flytekit.Literal, expected_python_type: Type[ShuffleJob]
    ) -> ShuffleJob:
        """
        In this function we want to be able to re-hydrate the custom object from Flyte Literal value
        """
        # Step 1: lets download remote data locally
        local_path = ctx.file_access.get_random_local_path()
        ctx.file_access.download(lv.scalar.blob.uri, local_path)
        # Step 2: create the ShuffleJob object
        with open(local_path, "rb") as f:
            return self._shuffle_jobs_from_bytes(f.read())


@dataclasses.dataclass
class ShuffleWorkerResults:
    partition: str
    pa_table: pa.Table


class ShuffleWorkerResultsTransformer(TypeTransformer[ShuffleWorkerResults]):
    _TYPE_INFO = flytekit.BlobType(
        format="binary",
        dimensionality=flytekit.BlobType.BlobDimensionality.SINGLE,
    )

    def __init__(self):
        super(ShuffleWorkerResultsTransformer, self).__init__(
            name="shuffleworkerresults-transform",
            t=ShuffleWorkerResults,
        )

    def get_literal_type(self, t: Type[ShuffleWorkerResults]) -> flytekit.LiteralType:
        """
        This is useful to tell the Flytekit type system that ``ShuffleWorkerResults`` actually refers to
        what corresponding type In this example, we say its of format binary (do not try to introspect) and
        there are more than one files in it
        """
        return flytekit.LiteralType(blob=self._TYPE_INFO)

    def to_literal(
        self,
        ctx: flytekit.FlyteContext,
        python_val: ShuffleWorkerResults,
        python_type: Type[ShuffleWorkerResults],
        expected: flytekit.LiteralType,
    ) -> flytekit.Literal:
        """
        This method is used to convert from given python type object ``ShuffleWorkerResults``
        to the Literal representation
        """
        # Step 1: lets upload all the data into a remote place recommended by Flyte
        remote_path = ctx.file_access.get_random_remote_path()
        local_path = ctx.file_access.get_random_local_path()
        with pa.ipc.new_stream(
            local_path, python_val.pa_table.schema.with_metadata({"partition": python_val.partition})
        ) as stream:
            stream.write(python_val.pa_table)
        ctx.file_access.upload(local_path, remote_path)
        # Step 2: lets return a pointer to this remote_path in the form of a literal
        return flytekit.Literal(
            scalar=flytekit.Scalar(
                blob=flytekit.Blob(uri=remote_path, metadata=flytekit.BlobMetadata(type=self._TYPE_INFO))
            )
        )

    def to_python_value(
        self, ctx: flytekit.FlyteContext, lv: flytekit.Literal, expected_python_type: Type[ShuffleWorkerResults]
    ) -> ShuffleWorkerResults:
        """
        In this function we want to be able to re-hydrate the custom object from Flyte Literal value
        """
        # Step 1: lets download remote data locally
        local_path = ctx.file_access.get_random_local_path()
        ctx.file_access.download(lv.scalar.blob.uri, local_path)
        # Step 2: create the ShuffleWorkerResults object
        with pa.ipc.open_stream(local_path) as reader:
            return ShuffleWorkerResults(
                pa_table=pa.Table.from_batches([b for b in reader]),
                partition=reader.schema.metadata[b"partition"].decode("utf-8"),
            )


TypeEngine.register(ShuffleJobTransformer())
TypeEngine.register(ShuffleWorkerResultsTransformer())


###
# Task and Workflow definitions
###


@flytekit.task(requests=flytekit.Resources(mem="2Gi", cpu="1"), retries=2)
def initialize_dataset(
    schema_json_str: str,
    dataset_id: str,
) -> str:
    """Write the schema to the storage."""
    s3_path_factory = S3PathFactory()
    s3_storage = S3DataStorage()
    schema_path = s3_path_factory.get_dataset_schema_path(DatasetID.from_str(dataset_id))
    s3_storage.put_object_s3(schema_json_str.encode("utf-8"), schema_path)
    return schema_json_str


@flytekit.task(requests=flytekit.Resources(mem="8Gi", cpu="2"), retries=2)
def create_shuffling_jobs(
    schema_json_str: str,
    dataset_id: str,
    worker_max_working_set_size: int = 16384,
) -> List[ShuffleJob]:
    """Read the DynamoDB and return a list of shuffling jobs to distribute.

    The job descriptions are stored into files managed by Flyte.
    :param dataset_id: string representation of the DatasetID we need to process (dataset name + version).
    :param schema_json_str: string representation of the dataset schema
    :param max_rows_per_worker: Maximum number of rows to assign per working set, defaults to 16384 but can be
        increased if dataset sizes are so large that we want to use fewer workers and don't mind the committing
        step taking longer per-worker.
    :return: a list of shuffling jobs to do.
    """
    # TODO(jchia): Dynamically decide on what backends to use for S3 and MetadataDatabase instead of hardcoding here
    backend = DatasetWriterBackend(S3PathFactory(), S3DataStorage(), DynamodbMetadataDatabase())
    job_factory = ShuffleJobFactory(backend, worker_max_working_set_size=worker_max_working_set_size)
    return list(job_factory.build_shuffle_jobs(DatasetID.from_str(dataset_id)))


@flytekit.task(requests=flytekit.Resources(mem="8Gi", cpu="2"), retries=4, cache=True, cache_version="v1")
def run_shuffling_job(job: ShuffleJob) -> ShuffleWorkerResults:
    """Run one shuffling job
    :param job: the ShuffleJob for this worker to run
    :return: pyarrow table containing only metadata and pointers to the ColumnBytesFiles in S3 for
        bytes columns in the dataset.
    """
    worker = ShuffleWorker(storage=S3DataStorage(), s3_path_factory=S3PathFactory())
    return ShuffleWorkerResults(
        pa_table=worker.process_job(job),
        partition=job.dataset_partition.partition,
    )


@flytekit.task(requests=flytekit.Resources(mem="8Gi", cpu="2"), retries=2)
def finalize_shuffling_jobs(dataset_id: str, shuffle_results: List[ShuffleWorkerResults]) -> Dict[str, int]:
    """Aggregate the indexes from the various shuffling jobs and publish the final parquet files for the dataset.
    :param dataset_id: string representation of the DatasetID we need to process (dataset name + version).
    :param shuffled_jobs_files: list of files containing the pandas Dataframe generated by the shuffling jobs.
    :return: A dictionary mapping partition_name -> size_of_partition.
    """
    results_by_partition = collections.defaultdict(list)
    for result in shuffle_results:
        results_by_partition[result.partition].append(result.pa_table)

    tables_by_partition: Dict[str, pa.Table] = {}
    for partition in results_by_partition:
        tables_by_partition[partition] = pa.concat_tables(results_by_partition[partition])

    dataset_id_obj = DatasetID.from_str(dataset_id)
    save_index(
        dataset_id_obj.name,
        dataset_id_obj.version,
        tables_by_partition,
        s3_storage=S3DataStorage(),
        s3_path_factory=S3PathFactory(),
    )
    return {partition_name: len(tables_by_partition[partition_name]) for partition_name in tables_by_partition}


@flytekit.workflow  # type: ignore
def WickerDataShufflingWorkflow(
    dataset_id: str,
    schema_json_str: str,
    worker_max_working_set_size: int = 16384,
) -> Dict[str, int]:
    """Pipeline finalizing a wicker dataset.
    :param dataset_id: string representation of the DatasetID we need to process (dataset name + version).
    :param schema_json_str: string representation of the schema, serialized as JSON
    :return: A dictionary mapping partition_name -> size_of_partition.
    """
    schema_json_str_committed = initialize_dataset(
        dataset_id=dataset_id,
        schema_json_str=schema_json_str,
    )
    jobs = create_shuffling_jobs(
        schema_json_str=schema_json_str_committed,
        dataset_id=dataset_id,
        worker_max_working_set_size=worker_max_working_set_size,
    )
    shuffle_results = flytekit.map_task(run_shuffling_job, metadata=flytekit.TaskMetadata(retries=1))(job=jobs)
    result = cast(Dict[str, int], finalize_shuffling_jobs(dataset_id=dataset_id, shuffle_results=shuffle_results))
    return result



from __future__ import annotations

import abc
import contextlib
import dataclasses
import hashlib
import os
import pickle
import threading
import time
from concurrent.futures import Executor, Future, ThreadPoolExecutor
from types import TracebackType
from typing import Any, Dict, Generator, Iterator, List, Optional, Tuple, Type, Union

from wicker.core.definitions import DatasetDefinition, DatasetID
from wicker.core.errors import WickerDatastoreException
from wicker.core.storage import S3DataStorage, S3PathFactory
from wicker.schema import dataparsing, serialization


@dataclasses.dataclass
class ExampleKey:
    """Unique identifier for one example."""

    # Name of the partition (usually train/eval/test)
    partition: str
    # Values of the primary keys for the example (in order of key precedence)
    primary_key_values: List[Union[str, int]]

    def hash(self) -> str:
        return hashlib.sha256(
            "/".join([self.partition, *[str(obj) for obj in self.primary_key_values]]).encode("utf-8")
        ).hexdigest()


@dataclasses.dataclass
class MetadataDatabaseScanRow:
    """Container for data obtained by scanning the MetadataDatabase"""

    partition: str
    row_data_path: str
    row_size: int


class AbstractDatasetWriterMetadataDatabase:
    """Database for storing metadata, used from inside the DatasetWriterBackend.

    NOTE: Implementors - this is the main implementation integration point for creating a new kind of DatasetWriter.
    """

    @abc.abstractmethod
    def save_row_metadata(self, dataset_id: DatasetID, key: ExampleKey, location: str, row_size: int) -> None:
        """Saves a row in the metadata database, marking it as having been uploaded to S3 and
        ready for shuffling.

        :param dataset_id: The ID of the dataset to save to
        :param key: The key of the example
        :param location: The location of the example in S3
        :param row_size: The size of the file in S3
        """
        pass

    @abc.abstractmethod
    def scan_sorted(self, dataset_id: DatasetID) -> Generator[MetadataDatabaseScanRow, None, None]:
        """Scans the MetadataDatabase for a **SORTED** stream of MetadataDatabaseScanRows for a given dataset.
        The stream is sorted by partition first, and then primary_key_values second.

        Should be fast O(minutes) to perform as this will be called from a single machine to assign chunks to jobs
        to run.

        :param dataset_id: The dataset to scan the metadata database for
        :return: a Generator of ExampleDBRows in **SORTED** partition + primary_key order
        """
        pass


class DatasetWriterBackend:
    """The backend for a DatasetWriter.

    Responsible for saving and retrieving data used during the dataset writing and committing workflow.
    """

    def __init__(
        self,
        s3_path_factory: S3PathFactory,
        s3_storage: S3DataStorage,
        metadata_database: AbstractDatasetWriterMetadataDatabase,
    ):
        self._s3_path_factory = s3_path_factory
        self._s3_storage = s3_storage
        self._metadata_db = metadata_database

    def save_row(self, dataset_id: DatasetID, key: ExampleKey, raw_data: Dict[str, Any]) -> None:
        """Adds an example to the backend

        :param dataset_id: ID of the dataset to save the row to
        :param key: Key of the example to write
        :param raw_data: raw data for the example that conforms to the schema provided at initialization
        """
        hashed_row_key = key.hash()
        pickled_row = pickle.dumps(raw_data)  # TODO(jchia): Do we want a more sophisticated storage format here?
        row_s3_path = os.path.join(
            self._s3_path_factory.get_temporary_row_files_path(dataset_id),
            hashed_row_key,
        )

        # Persist data in S3 and in MetadataDatabase
        self._s3_storage.put_object_s3(pickled_row, row_s3_path)
        self._metadata_db.save_row_metadata(dataset_id, key, row_s3_path, len(pickled_row))

    def commit_schema(
        self,
        dataset_definition: DatasetDefinition,
    ) -> None:
        """Write the schema to the backend as part of the commit step."""
        schema_path = self._s3_path_factory.get_dataset_schema_path(dataset_definition.identifier)
        serialized_schema = serialization.dumps(dataset_definition.schema)
        self._s3_storage.put_object_s3(serialized_schema.encode(), schema_path)


DEFAULT_BUFFER_SIZE_LIMIT = 1000


class DatasetWriter:
    """DatasetWriter providing async writing functionality. Implementors should override
    the ._save_row_impl method to define functionality for saving each individual row from inside the
    async thread executors.
    """

    def __init__(
        self,
        dataset_definition: DatasetDefinition,
        metadata_database: AbstractDatasetWriterMetadataDatabase,
        s3_path_factory: Optional[S3PathFactory] = None,
        s3_storage: Optional[S3DataStorage] = None,
        buffer_size_limit: int = DEFAULT_BUFFER_SIZE_LIMIT,
        executor: Optional[Executor] = None,
        wait_flush_timeout_seconds: int = 300,
    ):
        """Create a new DatasetWriter

        :param dataset_definition: definition of the dataset
        :param s3_path_factory: factory for s3 paths
        :param s3_storage: S3-compatible storage for storing data
        :param buffer_size_limit: size limit to the number of examples buffered in-memory, defaults to 1000
        :param executor: concurrent.futures.Executor to use for async writing, defaults to None
        :param wait_flush_timeout_seconds: number of seconds to wait before timing out on flushing
            all examples, defaults to 10
        """
        self.dataset_definition = dataset_definition
        self.backend = DatasetWriterBackend(
            s3_path_factory if s3_path_factory is not None else S3PathFactory(),
            s3_storage if s3_storage is not None else S3DataStorage(),
            metadata_database,
        )

        self.buffer: List[Tuple[ExampleKey, Dict[str, Any]]] = []
        self.buffer_size_limit = buffer_size_limit

        self.wait_flush_timeout_seconds = wait_flush_timeout_seconds
        self.executor = executor if executor is not None else ThreadPoolExecutor(max_workers=16)
        self.writes_in_flight: Dict[str, Dict[str, Any]] = {}
        self.flush_condition_variable = threading.Condition()

    def __enter__(self) -> DatasetWriter:
        return self

    def __exit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.flush(block=True)

    def __del__(self) -> None:
        self.flush(block=True)

    def add_example(self, partition_name: str, raw_data: Dict[str, Any]) -> None:
        """Adds an example to the writer

        :param partition_name: partition name where the example belongs
        :param raw_data: raw data for the example that conforms to the schema provided at initialization
        """
        # Run sanity checks on the data, fill empty fields.
        ex = dataparsing.parse_example(raw_data, self.dataset_definition.schema)
        example_key = ExampleKey(
            partition=partition_name, primary_key_values=[ex[k] for k in self.dataset_definition.schema.primary_keys]
        )
        self.buffer.append((example_key, ex))

        # Flush buffer to persistent storage if at size limit
        if len(self.buffer) > self.buffer_size_limit:
            self.flush(block=False)

    def flush(self, block: bool = True) -> None:
        """Flushes the writer

        :param block: whether to block on flushing all currently buffered examples, defaults to True
        :raises TimeoutError: timing out on flushing all examples
        """
        batch_data = self.buffer
        self.buffer = []
        self._save_batch_data(batch_data)

        if block:
            with self._block_on_writes_in_flight(max_in_flight=0, timeout_seconds=self.wait_flush_timeout_seconds):
                pass

    @contextlib.contextmanager
    def _block_on_writes_in_flight(self, max_in_flight: int = 0, timeout_seconds: int = 60) -> Iterator[None]:
        """Blocks until number of writes in flight <= max_in_flight and yields control to the with block

        Usage:
        >>> with self._block_on_writes_in_flight(max_in_flight=10):
        >>>     # Code here holds the exclusive lock on self.flush_condition_variable
        >>>     ...
        >>> # Code here has released the lock, and has no guarantees about the number of writes in flight

        :param max_in_flight: maximum number of writes in flight, defaults to 0
        :param timeout_seconds: maximum number of seconds to wait, defaults to 10
        :raises TimeoutError: if waiting for more than timeout_seconds
        """
        start_time = time.time()
        with self.flush_condition_variable:
            while len(self.writes_in_flight) > max_in_flight:
                if time.time() - start_time > timeout_seconds:
                    raise TimeoutError(
                        f"Timed out while flushing dataset writes with {len(self.writes_in_flight)} writes in flight"
                    )
                self.flush_condition_variable.wait()
            yield

    def _save_row(self, key: ExampleKey, data: Dict[str, Any]) -> str:
        self.backend.save_row(self.dataset_definition.identifier, key, data)
        return key.hash()

    def _save_batch_data(self, batch_data: List[Tuple[ExampleKey, Dict[str, Any]]]) -> None:
        """Save a batch of data to persistent storage

        :param row_keys: Unique identifiers for each row Uses only 0-9A-F, can be used as a file name.
        :param batch_data: Batch of data
        """

        def done_callback(future: Future[str]) -> None:
            # TODO(jchia): We can add retries on failure by reappending to the buffer
            # this currently may raise a CancelledError or some Exception thrown by the .save_row function
            with self.flush_condition_variable:
                del self.writes_in_flight[future.result()]
                self.flush_condition_variable.notify()

        for key, data in batch_data:
            # Keep the number of writes in flight always smaller than 2 * self.buffer_size_limit
            with self._block_on_writes_in_flight(
                max_in_flight=2 * self.buffer_size_limit,
                timeout_seconds=self.wait_flush_timeout_seconds,
            ):
                key_hash = key.hash()
                if key_hash in self.writes_in_flight:
                    raise WickerDatastoreException(
                        f"Error: data example has non unique key {key}, primary keys must be unique"
                    )
                self.writes_in_flight[key_hash] = data
            future = self.executor.submit(self._save_row, key, data)
            future.add_done_callback(done_callback)



import os
from typing import Any, Dict, Literal

import wandb

from wicker.core.config import get_config
from wicker.core.definitions import DatasetID
from wicker.core.storage import S3PathFactory


def version_dataset(
    dataset_name: str,
    dataset_version: str,
    entity: str,
    metadata: Dict[str, Any],
    dataset_backend: Literal["s3"] = "s3",
) -> None:
    """
    Version the dataset on Weights and Biases using the config parameters defined in wickerconfig.json.

    Args:
        dataset_name: The name of the dataset to be versioned
        dataset_version: The version of the dataset to be versioned
        entity: Who the run will belong to
        metadata: The metadata to be logged as an artifact, enforces dataclass for metadata documentation
        dataset_backend: The backend where the dataset is stored, currently only supports s3
    """
    # needs to first acquire and set wandb creds
    # WANDB_API_KEY, WANDB_BASE_URL
    # _set_wandb_credentials()

    # needs to init the wandb run, this is going to be a 'data' run
    dataset_run = wandb.init(project="dataset_curation", name=f"{dataset_name}_{dataset_version}", entity=entity)

    # grab the uri of the dataset to be versioned
    dataset_uri = _identify_s3_url_for_dataset_version(dataset_name, dataset_version, dataset_backend)

    # establish the artifact and save the dir/s3_url to the artifact
    data_artifact = wandb.Artifact(f"{dataset_name}_{dataset_version}", type="dataset")
    data_artifact.add_reference(dataset_uri, name="dataset")

    # save metadata dict to the artifact
    data_artifact.metadata["version"] = dataset_version
    data_artifact.metadata["s3_uri"] = dataset_uri
    for key, value in metadata.items():
        data_artifact.metadata[key] = value

    # save the artifact to the run
    dataset_run.log_artifact(data_artifact)  # type: ignore
    dataset_run.finish()  # type: ignore


def _set_wandb_credentials() -> None:
    """
    Acquire the weights and biases credentials and load them into the environment.

    This load the variables into the environment as ENV Variables for WandB to use,
    this function overrides the previously set wandb env variables with the ones specified in
    the wicker config if they exist.
    """
    # load the config
    config = get_config()

    # if the keys are present in the config add them to the env
    wandb_config = config.wandb_config
    for field in wandb_config.__dataclass_fields__:  # type: ignore
        attr = wandb_config.__getattribute__(field)
        if attr is not None:
            os.environ[str(field).upper()] = attr
        else:
            if str(field).upper() not in os.environ:
                raise EnvironmentError(
                    f"Cannot use W&B without setting {str(field.upper())}. "
                    f"Specify in either ENV or through wicker config file."
                )


def _identify_s3_url_for_dataset_version(
    dataset_name: str,
    dataset_version: str,
    dataset_backend: Literal["s3"] = "s3",
) -> str:
    """
    Identify and return the s3 url for the dataset and version specified in the backend.

    Args:
        dataset_name: name of the dataset to retrieve url
        dataset_version: version of the dataset to retrieve url
        dataset_backend: backend of the dataset to retrieve url

    Returns:
        The url pointing to the dataset on storage
    """
    schema_path = ""
    if dataset_backend == "s3":
        # needs to do the parsing work to fetch the correct s3 uri
        schema_path = S3PathFactory().get_dataset_assets_path(DatasetID(name=dataset_name, version=dataset_version))
    return schema_path
